"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007
    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from collections import defaultdict

import geopandas as gpd
import networkx as nx
import pandas as pd
from shapely.geometry import LineString, MultiLineString, MultiPoint, Point
from shapely.ops import linemerge
from snkit.network import Network as SnkitNetwork
from tqdm import tqdm

NxGraph = nx.Graph | nx.MultiGraph | nx.MultiDiGraph


"""
Disclaimer!

This file contains several complex logic introduced in feature #277.
At the moment it was not possible to streamline / improve it further than
its current state.
"""


def merge_edges(
    snkit_network: SnkitNetwork,
    aggregate_func: str | dict,
    by: str | list,
    id_col: str,
) -> SnkitNetwork:
    """
    Merges the edges of a given `snkit.network.Network`.

    Args:
        snkit_network (SnkitNetwork): network to merge.
        aggregate_func (str | dict): Aggregation function to apply.
        by (str | list): Arguments (column names).
        id_col (str, optional): Name of the column representing the 'id'.

    Returns:
        SnkitNetwork: _description_
    """

    def _node_connectivity_degree(node, snkit_network: SnkitNetwork) -> int:
        return len(
            snkit_network.edges[
                (snkit_network.edges.from_id == node)
                | (snkit_network.edges.to_id == node)
            ]
        )

    def _get_edge_ids_to_update(edges_list: list) -> list:
        ids_to_update = []
        for edges in edges_list:
            ids_to_update.extend(edges.id.tolist())
        return ids_to_update

    def _get_merged_edges(
        paths_to_group: list,
        by: list,
        aggfunc: str | dict,
        net: SnkitNetwork,
    ) -> gpd.GeoDataFrame:
        updated_edges = gpd.GeoDataFrame(
            columns=net.edges.columns, crs=net.edges.crs
        )  # merged edges

        for edge_path in tqdm(paths_to_group, desc="merge_edge_paths"):
            # Convert None values to a placeholder value
            placeholder = "None"
            for col in by:
                edge_path[col] = edge_path[col].fillna(placeholder)
            merged_edges = _get_merge_edge_paths(edge_path, by, aggfunc, net)
            updated_edges = pd.concat([updated_edges, merged_edges], ignore_index=True)

        updated_edges_gdf = gpd.GeoDataFrame(updated_edges, geometry="geometry")
        updated_edges_gdf.set_crs(net.edges.crs, inplace=True)
        updated_edges_gdf = updated_edges_gdf.drop(columns=["id"])
        return updated_edges_gdf

    def _get_edge_paths(node_set: set, snkit_network: SnkitNetwork) -> list:
        # Convert edges to an adjacency list using vectorized operations
        edge_dict = defaultdict(set)
        from_ids = snkit_network.edges["from_id"].values
        to_ids = snkit_network.edges["to_id"].values

        for from_id, to_id in zip(from_ids, to_ids):
            edge_dict[from_id].add(to_id)
            edge_dict[to_id].add(from_id)

        edge_paths = []

        while node_set:
            popped_node = node_set.pop()
            node_path = {popped_node}
            candidates = {popped_node}
            while candidates:
                popped_cand = candidates.pop()
                matches = edge_dict[popped_cand]
                matches = matches - node_path
                for match in matches:
                    if match in node_set:
                        candidates.add(match)
                        node_path.add(match)
                        node_set.remove(match)
                    else:
                        node_path.add(match)
            if len(node_path) >= 2:
                edge_paths.append(
                    snkit_network.edges.loc[
                        (snkit_network.edges.from_id.isin(node_path))
                        & (snkit_network.edges.to_id.isin(node_path))
                    ]
                )
        return edge_paths

    if "degree" not in snkit_network.nodes.columns:
        snkit_network.nodes["degree"] = snkit_network.nodes[id_col].apply(
            lambda x: _node_connectivity_degree(x, snkit_network)
        )

    degree_2 = list(snkit_network.nodes[id_col].loc[snkit_network.nodes.degree == 2])
    degree_2_set = set(degree_2)
    edge_paths = _get_edge_paths(degree_2_set, snkit_network)

    edge_ids_to_update = _get_edge_ids_to_update(edge_paths)
    edges_to_keep = snkit_network.edges[
        ~snkit_network.edges["id"].isin(edge_ids_to_update)
    ]

    updated_edges = _get_merged_edges(
        paths_to_group=edge_paths,
        by=by,
        aggfunc=aggregate_func,
        net=snkit_network,
    )
    edges_to_keep = edges_to_keep.drop(columns=["id"])
    updated_edges = updated_edges.reset_index(drop=True)

    new_edges = pd.concat([edges_to_keep, updated_edges], ignore_index=True)
    new_edges_gdf = gpd.GeoDataFrame(new_edges, geometry="geometry")
    new_edges_gdf.set_crs(edges_to_keep.crs, inplace=True)
    new_edges_gdf = new_edges_gdf.reset_index(drop=True)

    nodes_to_keep = list(set(new_edges.from_id.tolist() + new_edges.to_id.tolist()))
    new_nodes_gdf = snkit_network.nodes[snkit_network.nodes[id_col].isin(nodes_to_keep)]
    new_nodes_gdf = new_nodes_gdf.reset_index(drop=True)

    merged_snkit_network = SnkitNetwork(nodes=new_nodes_gdf, edges=new_edges_gdf)
    merged_snkit_network.nodes["degree"] = merged_snkit_network.nodes[id_col].apply(
        lambda x: _node_connectivity_degree(x, merged_snkit_network)
    )

    return merged_snkit_network


def _merge_connected_lines(
    gdf: gpd.GeoDataFrame, by: str, aggfunc: dict
) -> gpd.GeoDataFrame:
    """
    Merge connected lines in a GeoDataFrame into a single LineString.

    Parameters:
    gdf (gpd.GeoDataFrame): GeoDataFrame containing the lines to merge.
    by (str): Column name to group by.
    aggfunc (dict): Dictionary of aggregation functions for other columns.

    Returns:
    gpd.GeoDataFrame: GeoDataFrame with merged lines.
    """

    # Merge all geometries into a single MultiLineString
    merged_geometry = linemerge(gdf.geometry.tolist())
    indices = gdf[by].iloc[0]
    # Create a new GeoDataFrame with the merged geometry
    merged_gdf = gpd.GeoDataFrame(
        [{**indices, "geometry": merged_geometry}], crs=gdf.crs
    )
    merged_gdf.set_index(by, inplace=True)

    # Combine the attributes using the aggregation function
    for col, func in aggfunc.items():
        if col != "geometry":
            # Try to convert the column to float if needed
            if gdf[col].dtype != float:
                try:
                    gdf[col] = gdf[col].astype(float)
                except ValueError:
                    pass  # Skip conversion if it fails
            merged_gdf[col] = [func(gdf[col])]

    return merged_gdf


def _get_merge_edge_paths(
    edges: gpd.GeoDataFrame,
    excluded_edge_types: list,
    aggfunc: str | dict,
    snkit_network: SnkitNetwork,
) -> gpd.GeoDataFrame:
    def get_connected_lines(ids: pd.Index):
        """
        Find groups of connected lines in a GeoDataFrame.

        Parameters:
        gdf (GeoDataFrame): A GeoDataFrame containing LINESTRING geometries.

        Returns:
        list of lists: Each sublist contains indices of lines in gdf that are connected.
        """

        # Initialize an empty graph
        _networkx_graph = nx.Graph()
        gdf = edges.loc[ids.tolist()]
        # Add edges to the graph for each line in the GeoDataFrame
        for idx, row in gdf.iterrows():
            # Get the start and end points of the line
            line = row["geometry"]
            start_point = line.coords[0]
            end_point = line.coords[-1]

            # Add the line as an edge between its start and end points with the id as attribute
            _networkx_graph.add_edge(start_point, end_point, index=idx, id=row["id"])

        # Find connected components in the graph
        connected_components = list(nx.connected_components(_networkx_graph))

        # Map each component to the corresponding line ids
        connected_line_groups = []
        for component in connected_components:
            line_ids = []
            for _, _, data in _networkx_graph.edges(component, data=True):
                line_ids.append(data["id"])
            connected_line_groups.append(line_ids)

        return connected_line_groups

    def _get_paths_to_merge(groups: dict) -> list:
        _paths_to_merge = []  # list of gpds to merge
        for _, edge_group_ids in groups.items():
            sub_path_parts = get_connected_lines(edge_group_ids)
            _paths_to_merge.extend(sub_path_parts)
        return _paths_to_merge

    # _get_merged_paths starts here
    grouped_edges = edges.groupby(excluded_edge_types)
    if len(grouped_edges.groups) == 1:
        merged_edges = GdfSnkitNetworkMerger(
            geo_dataframe=edges, snkit_network=snkit_network
        ).merge(
            by=excluded_edge_types,
            aggregate_func=aggfunc,
        )
    else:
        merged_edges = gpd.GeoDataFrame(
            columns=edges.columns, crs=edges.crs
        )  # merged edges
        edge_groups = edges.groupby(excluded_edge_types).groups
        paths_to_merge = _get_paths_to_merge(edge_groups)

        for path_ids in paths_to_merge:
            path_to_merge = edges[
                edges["id"].isin(path_ids)
            ].copy()  # indices of the edges in edges gdf
            merged = GdfSnkitNetworkMerger(
                geo_dataframe=path_to_merge, snkit_network=snkit_network
            ).merge(
                by=excluded_edge_types,
                aggregate_func=aggfunc,
            )
            merged_edges = pd.concat([merged_edges, merged], ignore_index=True)

        merged_edges.crs = edges.crs

    return merged_edges


class GdfSnkitNetworkMerger:
    """
    Merger of a `gpd.GeoDataFrame` and a `snkit.network.network`.
    This class was created to contain the related close and try reducing the code's complexity.
    """

    def __init__(
        self, geo_dataframe: gpd.GeoDataFrame, snkit_network: SnkitNetwork
    ) -> None:
        self._geo_dataframe = geo_dataframe
        self._snkit_network = snkit_network

    def merge(
        self,
        by: list,
        aggregate_func: dict,
    ) -> gpd.GeoDataFrame:
        """
        Merges the inner defined `gpd.GeoDataFrame` and `snkit.network.Network`
        based on the given arguments (`by`) and aggregation function ()`aggregate_func`).
        """
        _geo_dataframe = self._geo_dataframe
        _snkit_network = self._snkit_network
        # _merge function starts from here:
        self._geo_dataframe["intersections"] = _geo_dataframe.apply(
            lambda x: self._get_intersections(x, _geo_dataframe), axis=1
        )
        # _merged = gdf.dissolve(by=by, aggfunc=_aggfunc, sort=False)
        _merged = _merge_connected_lines(_geo_dataframe, by, aggregate_func)
        if len(_geo_dataframe) == 1:
            # 1. no merging is occurring
            start_path_extremities = [_geo_dataframe.iloc[0]["from_id"]]
            end_path_extremities = [_geo_dataframe.iloc[0]["to_id"]]
            _merged.from_id = start_path_extremities[0]
            _merged.to_id = end_path_extremities[0]
        else:
            # 2. merging is occurring
            if (
                len(
                    _geo_dataframe[
                        _geo_dataframe["intersections"].apply(lambda x: len(x) == 1)
                    ]
                )
                == 0
            ):
                # 2.1. a loop with two nodes degree > 2
                gdf_node_ids = list(
                    set(_geo_dataframe.from_id.tolist() + _geo_dataframe.to_id.tolist())
                )
                gdf_node_slice = _snkit_network.nodes[
                    _snkit_network.nodes["id"].isin(gdf_node_ids)
                ]
                if len(gdf_node_slice[gdf_node_slice["degree"] > 2]) == 0:
                    # 2.1.1. a loop with only degree 2 edges => isolated from the rest of the graph
                    logging.warning(
                        """
                    A sub-graph loop isolated from the main graph is detected and removed.
                    This isolated part had %s nodes with node_fids {gdf_node_slice.id.tolist()} in
                    the input node graph.
                    """,
                        len(gdf_node_slice),
                    )
                    if "demand_edge" in _geo_dataframe.columns:
                        logging.warning(
                            """'This sub-graph had these demand nodes %s""",
                            (
                                _geo_dataframe[
                                    _geo_dataframe.demand_edge == 1
                                ].from_id.tolist()
                                + _geo_dataframe[
                                    _geo_dataframe.demand_edge == 1
                                ].to_id.tolist()
                            ),
                        )
                    return gpd.GeoDataFrame(
                        data=None,
                        columns=_snkit_network.edges.columns,
                        crs=_snkit_network.edges.crs,
                    )

                elif len(gdf_node_slice[gdf_node_slice["degree"] > 2]) == 1:
                    # 2.1.2. If there is only one node with the degree bigger than 2
                    if (
                        "demand_edge" not in _geo_dataframe.columns
                        or len(_geo_dataframe[_geo_dataframe["demand_edge"] == 1]) == 0
                    ):
                        # No demand node is in this loop. Then omit this loop and return empty gdf
                        return gpd.GeoDataFrame(
                            data=None,
                            columns=_snkit_network.edges.columns,
                            crs=_snkit_network.edges.crs,
                        )
                    elif (
                        "demand_edge" in _geo_dataframe.columns
                        and len(_geo_dataframe[_geo_dataframe["demand_edge"] == 1]) > 0
                    ):
                        demand_node_ids = [
                            i
                            for i in set(
                                _geo_dataframe.from_id.tolist()
                                + _geo_dataframe.to_id.tolist()
                            )
                            if (
                                _geo_dataframe[
                                    _geo_dataframe.demand_edge == 1
                                ].from_id.tolist()
                                + _geo_dataframe[
                                    _geo_dataframe.demand_edge == 1
                                ].to_id.tolist()
                            ).count(i)
                            == 2
                        ]
                        if len(demand_node_ids) > 1:
                            # merging this situation is skipped: not probable + complicated
                            return _geo_dataframe
                        else:
                            # Only one demand node exists in the loop
                            if isinstance(
                                linemerge(_merged.geometry.iloc[0]), MultiLineString
                            ):
                                # to exclude the merged geoms for which linemerge does not work
                                return _geo_dataframe
                            path_extremities_node_ids = {
                                x
                                for x in gdf_node_slice[
                                    gdf_node_slice["degree"] > 2
                                ].id.tolist()
                                + demand_node_ids
                            }
                            _merged = self._get_merged_multiple_demand_edges(
                                _merged, path_extremities_node_ids
                            )
                else:
                    # 2.1.3. the only remaining option is two nodes with degrees bigger than 2
                    if (
                        "demand_edge" not in _geo_dataframe.columns
                        or len(_geo_dataframe[_geo_dataframe["demand_edge"] == 1]) == 0
                    ):
                        # No demand node is in this loop. Then merge
                        _merged = self._get_merged_in_a_loop(_merged, gdf_node_slice)
                    else:
                        return _geo_dataframe
            else:
                # 2.2. merging non-loop paths
                path_extremities_node_ids = {
                    i
                    for i in set(
                        _geo_dataframe.from_id.tolist() + _geo_dataframe.to_id.tolist()
                    )
                    if (
                        _geo_dataframe.from_id.tolist() + _geo_dataframe.to_id.tolist()
                    ).count(i)
                    == 1
                }
                # if len(path_extremities_node_ids) > 0:
                if ("demand_edge" in _geo_dataframe.columns) and (
                    len(_geo_dataframe[_geo_dataframe["demand_edge"] == 1]) > 1
                ):
                    _merged = self._get_merged_multiple_demand_edges(
                        _merged, path_extremities_node_ids
                    )
                elif (
                    "demand_edge" in _geo_dataframe.columns
                    and len(_geo_dataframe[_geo_dataframe["demand_edge"] == 1]) <= 1
                ) or ("demand_edge" not in _geo_dataframe.columns):
                    # 2.2.2.no dem node is in the to_be_merged path or only one dem node. In the later case dem node
                    # will not be dissolved because it is in the path_extremities_node_ids
                    _merged = self._get_merged_one_or_none_demand_edges(
                        _merged, path_extremities_node_ids
                    )
                else:
                    raise ValueError(
                        f"""Check the lines with the following ids {_geo_dataframe.id.tolist()} """
                    )

        _merged.node_A = _merged.from_id
        _merged.node_B = _merged.to_id
        _merged.crs = _geo_dataframe.crs
        return _merged

    def _get_merged_in_a_loop(
        self, _merged: gpd.GeoDataFrame, gdf_node_slice: pd.DataFrame
    ) -> gpd.GeoDataFrame:
        # 2.1.2. pick one with one intersection point
        start_path_extrms = [gdf_node_slice[gdf_node_slice["degree"] > 2].iloc[0].id]
        end_path_extrms = [gdf_node_slice[gdf_node_slice["degree"] > 2].iloc[1].id]
        _merged.from_id = [
            start_path_extremity for start_path_extremity in start_path_extrms
        ]
        _merged.to_id = [end_path_extremity for end_path_extremity in end_path_extrms]
        return _merged

    def _get_merged_multiple_demand_edges(
        self, _merged: gpd.GeoDataFrame, path_extrms_nod_ids: set
    ) -> gpd.GeoDataFrame:
        def get_node_id(r: gpd.GeoSeries, attr: str, path_extrms_nod_ids: set) -> int:
            # to fill from_id and to_id of the to-be-merged paths
            if r[attr] == -1:
                for path_extremities_node_id in path_extrms_nod_ids:
                    path_extremities_node_geom = self._snkit_network.nodes[
                        self._snkit_network.nodes.id == path_extremities_node_id
                    ].geometry.iloc[0]
                    if r.geometry.intersects(path_extremities_node_geom):
                        return path_extremities_node_id
            else:
                return r[attr]

        _mrgd = self._get_split_edges_info(_merged)
        _mrgd.from_id = _mrgd.apply(
            lambda row: get_node_id(row, "from_id", path_extrms_nod_ids), axis=1
        )
        _mrgd.to_id = _mrgd.apply(
            lambda row: get_node_id(row, "to_id", path_extrms_nod_ids), axis=1
        )
        return _mrgd

    def _get_split_edges_info(self, merged: gpd.GeoDataFrame) -> tuple:
        # used for the cases where demand nodes exist in the to-be-merged paths
        # make the demand node from_id of the merged edge
        geo_dataframe = self._geo_dataframe
        dem_nod_ids = [
            i
            for i in set(geo_dataframe.from_id.tolist() + geo_dataframe.to_id.tolist())
            if (
                geo_dataframe[geo_dataframe.demand_edge == 1].from_id.tolist()
                + geo_dataframe[geo_dataframe.demand_edge == 1].to_id.tolist()
            ).count(i)
            == 2
        ]
        split_parts = [merged["geometry"].iloc[0]]
        split_edges_gdf = gpd.GeoDataFrame(columns=merged.columns)
        for dem_nod_id in dem_nod_ids:
            for part in split_parts:
                part_splits, split_edges_gdf = self._split(
                    merged, part, dem_nod_id, split_edges_gdf
                )
                if part_splits is not None:
                    split_parts.extend(part_splits)
                    split_parts.remove(part)
        return split_edges_gdf

    def _split(
        self,
        merged: gpd.GeoDataFrame,
        line_geom: MultiLineString | LineString,
        dem_nod_id: int,
        splits_gdf: gpd.GeoDataFrame,
    ) -> tuple:
        # used for the cases where demand nodes exist in the to-be-merged paths
        dem_nod_geom = self._snkit_network.nodes[
            self._snkit_network.nodes.id == dem_nod_id
        ].geometry.iloc[0]
        if line_geom.contains(dem_nod_geom):
            if isinstance(line_geom, MultiLineString):
                coords = [
                    linemerge(line_geom).coords[0],
                    linemerge(line_geom).coords[-1],
                ]
            else:
                coords = [line_geom.coords[0], line_geom.coords[-1]]
            # Add the coords from the points
            coords += dem_nod_geom.coords
            # Calculate the distance along the line for each point
            dists = [linemerge(line_geom).project(Point(p)) for p in coords]
            # sort the coordinates
            coords = [p for (d, p) in sorted(zip(dists, coords))]
            splits = [
                LineString([coords[i], coords[i + 1]]) for i in range(len(coords) - 1)
            ]
            splits_gdf = self._update_split_edges_gdf(
                merged, splits, dem_nod_id, splits_gdf, line_geom
            )
            return splits, splits_gdf
        else:
            return None, splits_gdf

    def _update_split_edges_gdf(
        self,
        merged: gpd.GeoDataFrame,
        parts: list,
        dem_nod_id: int,
        splt_edgs: gpd.GeoDataFrame,
        _split_line_geom: LineString,
    ) -> gpd.GeoDataFrame:
        # used for the cases where demand nodes exist in the to-be-merged paths
        for part in parts:
            # _split_line_geom is the line divided and produced parts
            if _split_line_geom not in splt_edgs.geometry.tolist():
                part_gdf = gpd.GeoDataFrame(
                    {
                        "geometry": part,
                        "id": len(splt_edgs),
                        "from_id": dem_nod_id,
                        "to_id": -1,
                        **merged.drop(columns=["geometry", "id", "from_id", "to_id"]),
                    }
                )
            else:
                # if _split_line_geom is divided and n stored in splt_edgs, we need to retrieve from/to_id info
                # and update splt_edgs
                part_gdf = gpd.GeoDataFrame(
                    {
                        "geometry": part,
                        "id": -1,
                        "from_id": splt_edgs[
                            splt_edgs.geometry == _split_line_geom
                        ].apply(
                            lambda row: (
                                row.from_id if row.from_id != -1 else dem_nod_id
                            ),
                            axis=1,
                        ),
                        "to_id": splt_edgs[
                            splt_edgs.geometry == _split_line_geom
                        ].apply(
                            lambda row: (row.to_id if row.to_id != -1 else dem_nod_id),
                            axis=1,
                        ),
                        **merged.drop(columns=["geometry", "id", "from_id", "to_id"]),
                    }
                )
                _split_line_index = splt_edgs.loc[
                    splt_edgs.geometry == _split_line_geom
                ].index[0]
                splt_edgs = splt_edgs.drop(_split_line_index)
            splt_edgs = pd.concat([splt_edgs, part_gdf], ignore_index=True)
        return splt_edgs

    def _get_merged_one_or_none_demand_edges(
        self, _merged, path_extrms_nod_ids: set
    ) -> gpd.GeoDataFrame:
        geo_dataframe = self._geo_dataframe
        _start_edges = geo_dataframe[
            geo_dataframe["intersections"].apply(lambda x: len(x) == 1)
        ]
        if ("demand_edge" in geo_dataframe.columns) and (
            len(geo_dataframe[geo_dataframe["demand_edge"] == 1])
        ) == 1:
            _start_edge = _start_edges[_start_edges.demand_edge == 1].iloc[0]
        elif ("demand_edge" not in geo_dataframe.columns) or (
            len(geo_dataframe[geo_dataframe["demand_edge"] == 1])
        ) != 1:
            _start_edge = _start_edges.iloc[0]
        start_path_extrms = [
            (
                _start_edge["from_id"]
                if _start_edge["from_id"] in list(path_extrms_nod_ids)
                else _start_edge["to_id"]
            )
        ]
        end_path_extrms = [(path_extrms_nod_ids - set(start_path_extrms)).pop()]
        _merged.from_id = [
            start_path_extremity for start_path_extremity in start_path_extrms
        ]
        _merged.to_id = [end_path_extremity for end_path_extremity in end_path_extrms]
        return _merged

    def _get_intersections(self, _edge, _edges):
        intersections = []
        edge_geometry = _edge.geometry.simplify(tolerance=1e-8)

        for _, other_edge in _edges.iterrows():
            other_edge_geometry = other_edge.geometry.simplify(tolerance=1e-8)

            if not edge_geometry.equals(other_edge_geometry):  # avoid self-intersection
                intersection = edge_geometry.intersection(other_edge_geometry)

                if not intersection.is_empty and any(
                    intersection.intersects(boundary)
                    for boundary in edge_geometry.boundary.geoms
                ):
                    if isinstance(intersection, MultiPoint):
                        intersections.extend(
                            [
                                point.coords[0]
                                for point in intersection.geoms
                                if point in other_edge_geometry.boundary.geoms
                            ]
                        )
                    else:
                        intersections.append(intersection.coords[0])

        return sorted(intersections, key=lambda x: x[0])



from __future__ import annotations

from ra2ce.configuration.ra2ce_enum_base import Ra2ceEnumBase


class SourceEnum(Ra2ceEnumBase):
    OSB_BPF = 1
    OSM_DOWNLOAD = 2
    SHAPEFILE = 3
    PICKLE = 4
    INVALID = 99

    @classmethod
    def get_enum(cls, input: str) -> SourceEnum:
        try:
            return cls[input.replace(" ", "_").upper()]
        except KeyError:
            return cls.INVALID

    @property
    def config_value(self) -> str:
        _parts = self.name.split("_")
        return " ".join(
            [_part if len(_part) == 3 else _part.lower() for _part in _parts]
        )



from ra2ce.configuration.ra2ce_enum_base import Ra2ceEnumBase


class RiskCalculationModeEnum(Ra2ceEnumBase):
    NONE = 0
    DEFAULT = 1
    CUT_FROM_YEAR = 2
    TRIANGLE_TO_NULL_YEAR = 3
    INVALID = 99



from ra2ce.configuration.ra2ce_enum_base import Ra2ceEnumBase


class NetworkTypeEnum(Ra2ceEnumBase):
    NONE = 0
    WALK = 1
    BIKE = 2
    DRIVE = 3
    DRIVE_SERVICE = 4
    ALL = 5
    INVALID = 99



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from dataclasses import dataclass

from networkx import MultiDiGraph


@dataclass
class ExtremitiesData:
    from_id: int = None
    to_id: int = None
    from_to_id: tuple = None
    to_from_id: tuple = None
    from_to_coor: tuple = None
    to_from_coor: tuple = None

    @staticmethod
    def get_extremities_data_for_sub_graph(
        from_node_id: int,
        to_node_id: int,
        sub_graph: MultiDiGraph,
        graph: MultiDiGraph,
        shared_elements: set,
    ):
        """Both extremities should be in the unique_graph still makes an edge between similar node to u (the node
        with u coordinates and different id, included in the unique_graph) and v Here, sub_graph is the unique_graph
        and graph is complex_graph Shared elements are shared btw sub_graph and graph, which are elements to include
        when dropping duplicates"""
        if shared_elements is None or not isinstance(shared_elements, set):
            raise ValueError("unique_elements should be a set")
        if from_node_id in sub_graph.nodes() and to_node_id in sub_graph.nodes():
            return ExtremitiesData.arrange_extremities_data(
                from_node_id=from_node_id, to_node_id=to_node_id, graph=sub_graph
            )
        elif (
            graph.nodes[from_node_id]["x"],
            graph.nodes[from_node_id]["y"],
        ) in shared_elements and to_node_id in sub_graph.nodes():

            from_node_id_prime = ExtremitiesData.find_node_id_by_coor(
                sub_graph,
                graph.nodes[from_node_id]["x"],
                graph.nodes[from_node_id]["y"],
            )
            if from_node_id_prime == to_node_id:
                return ExtremitiesData()
            else:
                return ExtremitiesData.arrange_extremities_data(
                    from_node_id=from_node_id_prime,
                    to_node_id=to_node_id,
                    graph=sub_graph,
                )

        elif (
            from_node_id in sub_graph.nodes()
            and (graph.nodes[to_node_id]["x"], graph.nodes[to_node_id]["y"])
            in shared_elements
        ):

            to_node_id_prime = ExtremitiesData.find_node_id_by_coor(
                sub_graph, graph.nodes[to_node_id]["x"], graph.nodes[to_node_id]["y"]
            )
            if from_node_id == to_node_id_prime:
                return ExtremitiesData()
            else:
                return ExtremitiesData.arrange_extremities_data(
                    from_node_id=from_node_id,
                    to_node_id=to_node_id_prime,
                    graph=sub_graph,
                )
        else:
            return ExtremitiesData()

    @staticmethod
    def arrange_extremities_data(
        from_node_id: int, to_node_id: int, graph: MultiDiGraph
    ):
        return ExtremitiesData(
            from_id=from_node_id,
            to_id=to_node_id,
            from_to_id=(from_node_id, to_node_id),
            to_from_id=(to_node_id, from_node_id),
            from_to_coor=(
                (graph.nodes[from_node_id]["x"], graph.nodes[to_node_id]["x"]),
                (graph.nodes[from_node_id]["y"], graph.nodes[to_node_id]["y"]),
            ),
            to_from_coor=(
                (graph.nodes[to_node_id]["x"], graph.nodes[from_node_id]["x"]),
                (graph.nodes[to_node_id]["y"], graph.nodes[from_node_id]["y"]),
            ),
        )

    @staticmethod
    def find_node_id_by_coor(graph: MultiDiGraph, target_x: float, target_y: float):
        """
        finds the node in unique graph with the same coor
        """
        for node, data in graph.nodes(data=True):
            if (
                "x" in data
                and "y" in data
                and data["x"] == target_x
                and data["y"] == target_y
            ):
                return node
        return None



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# -*- coding: utf-8 -*-

from __future__ import annotations

import logging
import warnings
from pathlib import Path
from typing import Optional

from shapely.errors import ShapelyDeprecationWarning

from ra2ce.analysis.analysis_config_data.analysis_config_data import AnalysisConfigData
from ra2ce.analysis.analysis_config_wrapper import AnalysisConfigWrapper
from ra2ce.analysis.analysis_result_wrapper import AnalysisResultWrapper
from ra2ce.configuration.config_factory import ConfigFactory
from ra2ce.configuration.config_wrapper import ConfigWrapper
from ra2ce.network.network_config_data.network_config_data import NetworkConfigData
from ra2ce.network.network_config_wrapper import NetworkConfigWrapper
from ra2ce.ra2ce_logger import Ra2ceLogger
from ra2ce.runners import AnalysisRunnerFactory

warnings.filterwarnings(
    action="ignore", message=".*initial implementation of Parquet.*"
)
warnings.filterwarnings(action="ignore", message="All-NaN slice encountered")
warnings.filterwarnings(action="ignore", message="Value *not successfully written.*")
warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
warnings.simplefilter(action="ignore", category=FutureWarning)


class Ra2ceHandler:
    """
    Top level class to handle the RA2CE analysis process.
    This class is used to orchestrate the analysis process based on the provided network and analysis configuration,
    including the logging configuration
    """

    input_config: ConfigWrapper

    def __init__(self, network: Optional[Path], analysis: Optional[Path]) -> None:
        if network or analysis:
            self._initialize_logger_from_files(network, analysis)
            self.input_config = ConfigFactory.get_config_wrapper(network, analysis)
        else:
            self._initialize_logger_from_config()

    @classmethod
    def from_config(
        cls, network: NetworkConfigData, analysis: AnalysisConfigData
    ) -> Ra2ceHandler:
        """
        Create a handler from the provided network and analysis configuration.

        Args:
            network (NetworkConfigData): Network configuration
            analysis (AnalysisConfigData): Analysis configuration

        Returns:
            Ra2ceHandler: The handler object
        """

        def set_config_paths(_analysis_config: AnalysisConfigWrapper) -> None:
            if not network:
                return
            _analysis_config.config_data.root_path = network.root_path
            _analysis_config.config_data.input_path = network.input_path
            _analysis_config.config_data.static_path = network.static_path
            _analysis_config.config_data.output_path = network.output_path

        def get_network_config() -> NetworkConfigWrapper | None:
            if not isinstance(network, NetworkConfigData):
                return None
            _network_config = NetworkConfigWrapper()
            _network_config.config_data = network
            if network.output_graph_dir:
                if network.output_graph_dir.is_dir():
                    _network_config.graph_files = (
                        _network_config.read_graphs_from_config(
                            network.output_graph_dir
                        )
                    )
                else:
                    network.output_graph_dir.mkdir(parents=True)
            return _network_config

        def get_analysis_config() -> AnalysisConfigWrapper | None:
            if not isinstance(analysis, AnalysisConfigData):
                return None
            _analysis_config = AnalysisConfigWrapper()
            _analysis_config.config_data = analysis
            set_config_paths(_analysis_config)
            if isinstance(_handler.input_config.network_config, NetworkConfigWrapper):
                _analysis_config.config_data.network = (
                    _handler.input_config.network_config.config_data.network
                )
                _analysis_config.config_data.origins_destinations = (
                    _handler.input_config.network_config.config_data.origins_destinations
                )
                _analysis_config.graph_files = (
                    _handler.input_config.network_config.graph_files
                )
            return _analysis_config

        _handler = cls(None, None)
        _handler.input_config = ConfigWrapper()
        _handler.input_config.network_config = get_network_config()
        _handler.input_config.analysis_config = get_analysis_config()

        return _handler

    def _initialize_logger_from_files(
        self, network: Optional[Path], analysis: Optional[Path]
    ) -> None:
        _output_config = None
        if network:
            _output_config = NetworkConfigData.get_data_output(network)
        elif analysis:
            _output_config = AnalysisConfigData.get_data_output(analysis)
        else:
            logging.warning(
                "No valid location provided to start logging: no logger and logfile is created."
            )
            return
        Ra2ceLogger.initialize_file_logger(
            logging_dir=_output_config, logger_name="RA2CE"
        )

    def _initialize_logger_from_config(self) -> None:
        Ra2ceLogger.initialize_console_logger(logger_name="RA2CE")

    def configure(self) -> None:
        """
        Configures the `ConfigWrapper` with the current `AnalysisConfigData` and
        `NetworkConfigData` so that the analyses can be succesfully run.
        """
        self.input_config.configure()

    def run_analysis(self) -> list[AnalysisResultWrapper]:
        """
        Runs a Ra2ce analysis based on the provided network and analysis files.

        Args: None

        Raises:
            ValueError: If the input files are not valid

        Returns:
            list[AnalysisResultWrapper]: A list of analysis results
        """
        if not self.input_config.analysis_config:
            return
        if not self.input_config.is_valid_input():
            _error = "Error validating input files. Ra2ce will close now."
            logging.error(_error)
            raise ValueError(_error)

        _runner = AnalysisRunnerFactory.get_runner(self.input_config)
        return _runner.run(self.input_config.analysis_config)

    @staticmethod
    def run_with_ini_files(
        network_ini_file: Path | None, analysis_ini_file: Path | None
    ) -> list[AnalysisResultWrapper]:
        """
        Streamlined method to directly run a `Ra2ce` analysis based
        on the provided `network` and `analysis` `.ini` files.

        This streamlined method allows for automatic initialization of the
        logger.

        Args:
            network_ini_file (Path | None): Location of the network file (`*.ini`).
            analysis_ini_file (Path | None): Location of the analysis file (`*.ini`).

        Returns:
            list[AnalysisResultWrapper]: A list of analyses results.
        """
        _handler = Ra2ceHandler(network_ini_file, analysis_ini_file)
        _handler.configure()
        return _handler.run_analysis()

    @staticmethod
    def run_with_config_data(
        network: NetworkConfigData | None, analysis: AnalysisConfigData | None
    ) -> list[AnalysisResultWrapper]:
        """
        Streamlined method to directly run a `Ra2ce` analysis based
        on the dataclasses for `Network` and `Analysis` instead of
        using `.ini` files.

        This streamlined method allows for automatic initialization of the
        logger.

        Args:
            network (NetworkConfigData | None):
                Dataclass containing all the information for the network.
            analysis (AnalysisConfigData | None):
                Dataclass containing all the information related to analyses.

        Returns:
            list[AnalysisResultWrapper]: A list of analyses results.
        """
        _handler = Ra2ceHandler.from_config(network, analysis)
        _handler.configure()
        return _handler.run_analysis()



import re
from configparser import ConfigParser
from pathlib import Path
from typing import Any, Union

from ra2ce.common.configuration.ini_configuration_reader_protocol import (
    ConfigDataReaderProtocol,
)
from ra2ce.network.network_config_data.enums.aggregate_wl_enum import AggregateWlEnum
from ra2ce.network.network_config_data.enums.network_type_enum import NetworkTypeEnum
from ra2ce.network.network_config_data.enums.road_type_enum import RoadTypeEnum
from ra2ce.network.network_config_data.enums.source_enum import SourceEnum
from ra2ce.network.network_config_data.network_config_data import (
    CleanupSection,
    HazardSection,
    IsolationSection,
    NetworkConfigData,
    NetworkSection,
    OriginsDestinationsSection,
    ProjectSection,
)


class NetworkConfigDataReader(ConfigDataReaderProtocol):
    _parser: ConfigParser

    def __init__(self) -> None:
        self._parser = ConfigParser(
            inline_comment_prefixes="#",
            converters={"list": lambda x: [x.strip() for x in re.split("[,|;]", x)]},
        )

    def read(self, ini_file: Path) -> NetworkConfigData:
        self._parser.read(ini_file)
        self._remove_none_values()

        _parent_dir = ini_file.parent

        _config_data = NetworkConfigData(
            root_path=_parent_dir.parent,
            input_path=_parent_dir.joinpath("input"),
            static_path=_parent_dir.joinpath("static"),
            output_path=_parent_dir.joinpath("output"),
            **self._get_sections(),
        )
        self._correct_paths(_config_data)
        return _config_data

    def _correct_paths(self, config_data: NetworkConfigData) -> None:
        """
        This method is created because we support defining Path properties with just their filename, but no relative or absolute references.
        These need to be done later on by combining the `input`, `static` or `output` paths. To avoid extra logic all over the solution,
        we can correct these paths here.

        Args:
            config_data (NetworkConfigData): The configuration data whose paths need to be corrected.
        """

        def _select_to_correct(path_value: Union[list[Path], Path]) -> bool:
            if not path_value:
                return False

            if isinstance(path_value, list):
                return _select_to_correct(path_value[0])
            return not path_value.exists()

        def _correct_list(path_root: Path, path_value_list: list[Path]) -> list[Path]:
            _corrected_list = []
            for _path_value in path_value_list:
                if not _path_value.exists():
                    _corrected_list.append(path_root.joinpath(_path_value))
                else:
                    _corrected_list.append(_path_value)
            return _corrected_list

        # Relative to network directory.
        _network_directory = config_data.static_path.joinpath("network")
        if _select_to_correct(config_data.origins_destinations.origins):
            config_data.origins_destinations.origins = _network_directory.joinpath(
                config_data.origins_destinations.origins
            )

        if _select_to_correct(config_data.origins_destinations.destinations):
            config_data.origins_destinations.destinations = _network_directory.joinpath(
                config_data.origins_destinations.destinations
            )

        if _select_to_correct(config_data.origins_destinations.region):
            config_data.origins_destinations.region = _network_directory.joinpath(
                config_data.origins_destinations.region
            )

        if _select_to_correct(config_data.network.polygon):
            config_data.network.polygon = _network_directory.joinpath(
                config_data.network.polygon
            )

        config_data.network.primary_file = _correct_list(
            _network_directory, config_data.network.primary_file
        )
        config_data.network.diversion_file = _correct_list(
            _network_directory, config_data.network.diversion_file
        )

        # Relative to hazard directory.
        _hazard_directory = config_data.static_path.joinpath("hazard")
        config_data.hazard.hazard_map = _correct_list(
            _hazard_directory, config_data.hazard.hazard_map
        )

    def _get_str_as_path(self, str_value: Union[str, Path]) -> Path:
        if str_value and not isinstance(str_value, Path):
            return Path(str_value)
        return str_value

    def _remove_none_values(self) -> None:
        # Remove 'None' from values, replace them with empty strings
        for _section in self._parser.sections():
            _keys_with_none = [
                k for k, v in self._parser[_section].items() if v == "None"
            ]
            for _key_with_none in _keys_with_none:
                self._parser[_section].pop(_key_with_none)

    def _get_sections(self) -> dict:
        return {
            "project": self.get_project_section(),
            "network": self.get_network_section(),
            "origins_destinations": self.get_origins_destinations_section(),
            "isolation": self.get_isolation_section(),
            "hazard": self.get_hazard_section(),
            "cleanup": self.get_cleanup_section(),
        }

    def get_project_section(self) -> ProjectSection:
        return ProjectSection(**self._parser["project"])

    def _get_path_list(
        self, section_name: str, property: str, fallback_opt: Any
    ) -> list[Path]:
        _value_list = self._parser.getlist(
            section_name, property, fallback=fallback_opt
        )
        return list(map(self._get_str_as_path, _value_list))

    def get_network_section(self) -> NetworkSection:
        _section = "network"
        _network_section = NetworkSection(**self._parser[_section])
        _network_section.source = SourceEnum.get_enum(
            self._parser.get(_section, "source", fallback=None)
        )
        _network_section.primary_file = self._get_path_list(
            _section, "primary_file", _network_section.primary_file
        )
        _network_section.diversion_file = self._get_path_list(
            _section, "diversion_file", _network_section.diversion_file
        )
        _network_section.directed = self._parser.getboolean(
            _section, "directed", fallback=_network_section.directed
        )
        _network_section.save_gpkg = self._parser.getboolean(
            _section, "save_gpkg", fallback=_network_section.save_gpkg
        )
        _network_section.network_type = NetworkTypeEnum.get_enum(
            self._parser.get(_section, "network_type", fallback=None)
        )
        _network_section.road_types = list(
            map(
                RoadTypeEnum.get_enum,
                self._parser.getlist(_section, "road_types", fallback=[]),
            )
        )
        _network_section.polygon = self._get_str_as_path(_network_section.polygon)
        _network_section.attributes_to_exclude_in_simplification = self._parser.getlist(
            _section, "attributes_to_exclude_in_simplification", fallback=[]
        )
        return _network_section

    def get_origins_destinations_section(self) -> OriginsDestinationsSection:
        _section = "origins_destinations"
        if _section not in self._parser:
            return OriginsDestinationsSection()

        _od_section = OriginsDestinationsSection(**self._parser[_section])
        _od_section.origin_out_fraction = self._parser.getint(
            _section, "origin_out_fraction", fallback=_od_section.origin_out_fraction
        )
        _od_section.origins = self._get_str_as_path(_od_section.origins)
        _od_section.destinations = self._get_str_as_path(_od_section.destinations)
        _od_section.region = self._get_str_as_path(_od_section.region)
        return _od_section

    def get_isolation_section(self) -> IsolationSection:
        _section = "isolation"
        if _section not in self._parser:
            return IsolationSection()

        return IsolationSection(**self._parser[_section])

    def get_hazard_section(self) -> HazardSection:
        _section = "hazard"
        if _section not in self._parser:
            return HazardSection()

        _hazard_section = HazardSection(**self._parser[_section])
        _hazard_section.hazard_map = list(
            map(
                self._get_str_as_path,
                self._parser.getlist(
                    _section, "hazard_map", fallback=_hazard_section.hazard_map
                ),
            )
        )
        _hazard_section.aggregate_wl = AggregateWlEnum.get_enum(
            self._parser.get(_section, "aggregate_wl", fallback=None)
        )
        _hazard_section.scenario_cost = list(
            self._parser.getlist(
                _section, "scenario_cost", fallback=_hazard_section.scenario_cost
            )
        )
        _hazard_section.overlay_segmented_network = self._parser.getboolean(
            _section,
            "overlay_segmented_network",
            fallback=_hazard_section.overlay_segmented_network,
        )
        return _hazard_section

    def get_cleanup_section(self) -> CleanupSection:
        _section = "cleanup"
        if _section not in self._parser:
            return CleanupSection()

        _cleanup_section = CleanupSection()
        _cleanup_section.snapping_threshold = self._parser.getboolean(
            _section,
            "snapping_threshold",
            fallback=_cleanup_section.snapping_threshold,
        )
        _cleanup_section.pruning_threshold = self._parser.getboolean(
            _section,
            "pruning_threshold",
            fallback=_cleanup_section.pruning_threshold,
        )
        _cleanup_section.segmentation_length = self._parser.getfloat(
            _section,
            "segmentation_length",
            fallback=_cleanup_section.segmentation_length,
        )
        _cleanup_section.merge_lines = self._parser.getboolean(
            _section, "merge_lines", fallback=_cleanup_section.merge_lines
        )
        _cleanup_section.cut_at_intersections = self._parser.getboolean(
            _section,
            "cut_at_intersections",
            fallback=_cleanup_section.cut_at_intersections,
        )
        _cleanup_section.delete_duplicate_nodes = self._parser.getboolean(
            _section,
            "delete_duplicate_nodes",
            fallback=_cleanup_section.delete_duplicate_nodes,
        )
        return _cleanup_section



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import math
from abc import ABC, abstractmethod
from pathlib import Path

import geopandas as gpd
import pandas as pd

from ra2ce.analysis.analysis_config_data.analysis_config_data import (
    AnalysisSectionLosses,
)
from ra2ce.analysis.analysis_config_data.enums.traffic_period_enum import (
    TrafficPeriodEnum,
)
from ra2ce.analysis.analysis_config_data.enums.trip_purpose_enum import TripPurposeEnum
from ra2ce.analysis.analysis_config_wrapper import AnalysisConfigWrapper
from ra2ce.analysis.analysis_input_wrapper import AnalysisInputWrapper
from ra2ce.analysis.losses.analysis_losses_protocol import AnalysisLossesProtocol
from ra2ce.analysis.losses.resilience_curves.resilience_curves_reader import (
    ResilienceCurvesReader,
)
from ra2ce.analysis.losses.time_values.time_values_reader import TimeValuesReader
from ra2ce.analysis.losses.traffic_intensities.traffic_intensities_reader import (
    TrafficIntensitiesReader,
)
from ra2ce.network.graph_files.graph_file import GraphFile
from ra2ce.network.hazard.hazard_names import HazardNames
from ra2ce.network.network_config_data.enums.aggregate_wl_enum import AggregateWlEnum
from ra2ce.network.network_config_data.enums.road_type_enum import RoadTypeEnum


class LossesBase(AnalysisLossesProtocol, ABC):
    """
    This class is the base class for the Losses analyses, containing the common methods and attributes.
    Based on the analysis type a different criticality analysis is executed.
    """

    analysis: AnalysisSectionLosses
    graph_file_hazard: GraphFile
    input_path: Path
    static_path: Path
    output_path: Path
    hazard_names: HazardNames

    def __init__(
        self,
        analysis_input: AnalysisInputWrapper,
        analysis_config: AnalysisConfigWrapper,
    ) -> None:
        self.analysis_input = analysis_input
        self.analysis_config = analysis_config
        self.analysis = analysis_input.analysis
        self.graph_file_hazard = analysis_input.graph_file_hazard

        self.link_id = analysis_config.config_data.network.file_id
        self.link_type_column = analysis_config.config_data.network.link_type_column
        self.trip_purposes: list[TripPurposeEnum] = self.analysis.trip_purposes

        self.performance_metric = f"diff_{self.analysis.weighing}"

        self.analysis_type = self.analysis.analysis
        self.traffic_period: TrafficPeriodEnum = self.analysis.traffic_period
        self.hours_per_traffic_period: float = self.analysis.hours_per_traffic_period
        if not self.hours_per_traffic_period:
            self.hours_per_traffic_period = 24
            if self.traffic_period != TrafficPeriodEnum.DAY:
                logging.warning(
                    "No value set for `hours_per_traffic_period`. Assuming 24 hours per traffic period."
                )

        self.production_loss_per_capita_per_hour = (
            self.analysis.production_loss_per_capita_per_hour
        )

        self._check_validity_analysis_files()
        self.intensities = TrafficIntensitiesReader(self.link_id).read(
            self.analysis.traffic_intensities_file
        )
        self.resilience_curves = ResilienceCurvesReader().read(
            self.analysis.resilience_curves_file
        )
        self.values_of_time = TimeValuesReader().read(self.analysis.values_of_time_file)

        self.input_path = analysis_input.input_path
        self.static_path = analysis_input.static_path
        self.output_path = analysis_input.output_path
        self.hazard_names = analysis_input.hazard_names

        self.result = gpd.GeoDataFrame()

    def _check_validity_analysis_files(self):
        if (
            self.analysis.traffic_intensities_file is None
            or self.analysis.resilience_curves_file is None
            or self.analysis.values_of_time_file is None
        ):
            raise ValueError(
                "traffic_intensities_file, resilience_curves_file, and values_of_time_file should be given"
            )

    def _get_disrupted_criticality_analysis_results(
        self, criticality_analysis: gpd.GeoDataFrame
    ):
        criticality_analysis.reset_index(inplace=True)

        if "key" in criticality_analysis.columns:
            criticality_analysis = criticality_analysis.drop_duplicates(
                ["u", "v", "key"]
            )
        else:
            criticality_analysis = criticality_analysis.drop_duplicates(["u", "v"])

        # filter out all links not affected by the hazard
        if self.analysis.aggregate_wl == AggregateWlEnum.NONE:
            self.criticality_analysis = criticality_analysis[
                criticality_analysis["EV1_ma"] > self.analysis.threshold
            ]
        elif self.analysis.aggregate_wl == AggregateWlEnum.MAX:
            self.criticality_analysis = criticality_analysis[
                criticality_analysis["EV1_max"] > self.analysis.threshold
            ]
        elif self.analysis.aggregate_wl == AggregateWlEnum.MEAN:
            self.criticality_analysis = criticality_analysis[
                criticality_analysis["EV1_mean"] > self.analysis.threshold
            ]
        elif self.analysis.aggregate_wl == AggregateWlEnum.MIN:
            self.criticality_analysis = criticality_analysis[
                criticality_analysis["EV1_min"] > self.analysis.threshold
            ]

        self.criticality_analysis_non_disrupted = criticality_analysis[
            ~criticality_analysis.index.isin(self.criticality_analysis.index)
        ]
        # link_id from list to tuple
        if len(self.criticality_analysis_non_disrupted) > 0:
            self.criticality_analysis_non_disrupted[
                self.link_id
            ] = self.criticality_analysis_non_disrupted[self.link_id].apply(
                lambda x: tuple(x) if isinstance(x, list) else x
            )
        self.criticality_analysis[self.link_id] = self.criticality_analysis[
            self.link_id
        ].apply(lambda x: tuple(x) if isinstance(x, list) else x)

        self.criticality_analysis.set_index(self.link_id, inplace=True)
        self.criticality_analysis_non_disrupted = (
            self.criticality_analysis_non_disrupted.reset_index()
        )

    def calculate_vehicle_loss_hours(self) -> gpd.GeoDataFrame:
        """
        This function opens an existing table with traffic data and value of time to calculate losses based on
        detouring values. It also includes a traffic jam estimation.
        """

        def _check_validity_criticality_analysis():
            if self.link_type_column not in self.criticality_analysis.columns:
                raise Exception(
                    f"""criticality_analysis results does not have the passed link_type_column.
            {self.link_type_column} is passed as link_type_column"""
                )

        def _get_range(height: float) -> tuple[float, float]:
            for range_tuple in _hazard_intensity_ranges:
                x, y = range_tuple
                if x <= height <= y:
                    return (x, y)
            raise ValueError(f"No matching range found for height {height}")

        def _create_result(vlh: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
            """

            Args: vlh: calculated vehicle_loss_hours GeoDataFrame. For single_link_losses it only includes the
            disrupted links. For Multi_link_losses it includes all links. This is because of the difference between
            the underlying single_link_redundancy and multi_link_redundancy analysis results.

            Returns: results of the Losses analysis. For the single_link_losses it adds non_disrupted links to vlh. For
            Multi_link_losses this is not necessary because of the underlying multi_link_redundancy analysis.

            """
            columns_without_index = [
                col
                for col in self.criticality_analysis_non_disrupted.columns
                if col not in ["level_0"]
            ]
            # Get the vlh_columns from vehicle_loss_hours that vlh calculations are filled in.
            vlh_columns = list(
                set(vlh.columns)
                - set(
                    self.criticality_analysis_non_disrupted[
                        columns_without_index
                    ].columns
                )
            )
            vlh[vlh_columns] = vlh[vlh_columns].fillna(0)

            result = pd.concat(
                [
                    vlh,
                    self.criticality_analysis_non_disrupted[columns_without_index],
                ]
            )
            result = result.reset_index()

            # Fill 0 for the vlh_columns of vlh and self.criticality_analysis_non_disrupted
            result.loc[result.index.difference(vlh.index), vlh_columns] = result.loc[
                result.index.difference(vlh.index), vlh_columns
            ].fillna(0)
            for col in ["index", "level_0"]:
                if col in result.columns:
                    result = result.drop(col, axis=1)

            return result

        _check_validity_criticality_analysis()

        _hazard_intensity_ranges = self.resilience_curves.ranges
        events = self.criticality_analysis.filter(regex=r"^EV(?!1_fr)")
        # Read the performance_change stating the functionality drop
        if "key" in self.criticality_analysis.columns:
            performance_change = self.criticality_analysis[
                [f"{self.performance_metric}", "u", "v", "key"]
            ]
            vehicle_loss_hours_df = pd.DataFrame(
                {
                    f"{self.link_id}": self.criticality_analysis.index.values,
                    "u": self.criticality_analysis["u"],
                    "v": self.criticality_analysis["v"],
                    "key": self.criticality_analysis["key"],
                }
            )
        else:
            performance_change = self.criticality_analysis[
                [f"{self.performance_metric}", "u", "v"]
            ]
            vehicle_loss_hours_df = pd.DataFrame(
                {
                    f"{self.link_id}": self.criticality_analysis.index.values,
                    "u": self.criticality_analysis["u"],
                    "v": self.criticality_analysis["v"],
                }
            )

        # shape vehicle_loss_hours
        # Check if the index name exists in the columns
        if vehicle_loss_hours_df.index.name in vehicle_loss_hours_df.columns:
            vehicle_loss_hours_df.reset_index(drop=True, inplace=True)
        else:
            vehicle_loss_hours_df.reset_index(inplace=True)

        # find the link_type and the hazard intensity
        connectivity_attribute = None
        if any(
            col in self.criticality_analysis.columns for col in ["detour", "connected"]
        ):
            connectivity_attribute = (
                "detour"
                if "detour" in self.criticality_analysis.columns
                else "connected"
            )
        vlh_additional_columns = self.criticality_analysis.columns.difference(
            vehicle_loss_hours_df.columns
        ).tolist()
        vehicle_loss_hours_df = pd.merge(
            vehicle_loss_hours_df,
            self.criticality_analysis[vlh_additional_columns],
            left_on=self.link_id,
            right_index=True,
        )
        vehicle_loss_hours = gpd.GeoDataFrame(
            vehicle_loss_hours_df,
            geometry="geometry",
            crs=self.criticality_analysis.crs,
        )
        for event in events.columns.tolist():
            for _, vlh_row in vehicle_loss_hours.iterrows():
                row_hazard_range = _get_range(vlh_row[event])
                row_connectivity = vlh_row[connectivity_attribute]
                row_performance_changes = performance_change.loc[
                    [vlh_row[self.link_id]]
                ]
                if "key" in vlh_row.index:
                    key = vlh_row["key"]
                else:
                    key = 0
                (u, v, k) = (
                    vlh_row["u"],
                    vlh_row["v"],
                    key,
                )
                # allow link_id not to be unique in the graph (results reliability is up to the user)
                # this can happen for instance when a directed graph should be made from an input network
                for performance_row in row_performance_changes.iterrows():
                    row_performance_change = performance_row[-1][
                        f"{self.performance_metric}"
                    ]
                    if "key" in performance_row[-1].index:
                        performance_key = performance_row[-1]["key"]
                    else:
                        performance_key = 0
                    row_u_v_k = (
                        performance_row[-1]["u"],
                        performance_row[-1]["v"],
                        performance_key,
                    )
                    if (
                        math.isnan(row_performance_change) and row_connectivity == 0
                    ) or row_performance_change == 0:
                        self._calculate_production_loss_per_capita(
                            vehicle_loss_hours, row_hazard_range, vlh_row, event
                        )
                    elif not (
                        math.isnan(row_performance_change)
                        and math.isnan(row_connectivity)
                    ) and ((u, v, k) == row_u_v_k):
                        self._populate_vehicle_loss_hour(
                            vehicle_loss_hours,
                            row_hazard_range,
                            vlh_row,
                            row_performance_change,
                            event,
                        )

        vehicle_loss_hours_result = _create_result(vehicle_loss_hours)
        return vehicle_loss_hours_result

    def _get_relevant_link_type(
        self, vlh_row: pd.Series, row_hazard_range: tuple[float, float]
    ) -> RoadTypeEnum:
        # Check if the resilience curve is present for the link type and hazard intensity
        _relevant_link_type = None
        if isinstance(vlh_row[self.link_type_column], list):
            # Find the link type with the highest disruption for the given hazard intensity
            _max_disruption = 0
            for _row_link_type in vlh_row[self.link_type_column]:
                _link_type = RoadTypeEnum.get_enum(_row_link_type)
                disruption = self.resilience_curves.get_disruption(
                    _link_type, row_hazard_range
                )
                if disruption > _max_disruption:
                    _relevant_link_type = _link_type
        else:
            _link_type = RoadTypeEnum.get_enum(vlh_row[self.link_type_column])
            if self.resilience_curves.has_resilience_curve(
                _link_type,
                row_hazard_range,
            ):
                _relevant_link_type = _link_type

        if not _relevant_link_type:
            raise ValueError(
                f"'{_link_type}' with range {row_hazard_range} was not found in the introduced resilience_curves"
            )

        return _relevant_link_type

    def _get_divisor(
        self, relevant_link_type: RoadTypeEnum, row_hazard_range: tuple[float, float]
    ):
        if all(
            ratio <= 1
            for ratio in self.resilience_curves.get_functionality_loss_ratio(
                relevant_link_type, row_hazard_range
            )
        ):
            return 1
        return 100  # high value assuming the road is almost inaccessible

    def _calculate_production_loss_per_capita(
        self,
        vehicle_loss_hours: gpd.GeoDataFrame,
        row_hazard_range: tuple[float, float],
        vlh_row: pd.Series,
        hazard_col_name: str,
    ):
        """
        In cases where there is no alternative route in the event of disruption of the road, we propose to use a
        proxy for the assessment of losses from the interruption of services from the road in these cases where no
        alternative routes exist.
        The assumption for the proxy is that a loss of production will occur from the
        interruption of the road, equal to the size of the added value from the persons that cannot make use of the
        road, measured in the regional GDP per capita. This assumption constitutes both the loss of production within
        the area that cannot be reached, as well the loss of production outside the area due to the inability of the
        workers from within the cut-off area to arrive at their place of production outside this area.

        The daily loss of productivity for each link section without detour routes, when they are
        disrupted is then obtained multiplying the traffic intensity by the total occupancy per vehicle type,
        including drivers, by the daily loss of productivity per capita per hour.

        The unit of time is hour.
        """
        vlh_total = 0
        _relevant_link_type = self._get_relevant_link_type(vlh_row, row_hazard_range)
        _divisor = self._get_divisor(_relevant_link_type, row_hazard_range)

        duration_steps = self.resilience_curves.get_duration_steps(
            _relevant_link_type, row_hazard_range
        )
        functionality_loss_ratios = self.resilience_curves.get_functionality_loss_ratio(
            _relevant_link_type, row_hazard_range
        )

        for trip_type in self.trip_purposes:
            intensity_trip_type = (
                self.intensities.calculate_intensity(
                    vlh_row[self.link_id], self.traffic_period, trip_type
                )
                / self.hours_per_traffic_period
            )

            occupancy_trip_type = self.values_of_time.get_occupants(trip_type)
            vlh_trip_type_event = sum(
                (
                    intensity_trip_type
                    * duration**2
                    * loss_ratio
                    * occupancy_trip_type
                    * self.production_loss_per_capita_per_hour
                )
                / _divisor
                for duration, loss_ratio in zip(
                    duration_steps, functionality_loss_ratios
                )
            )

            vehicle_loss_hours.loc[
                [vlh_row.name], f"vlh_{trip_type}_{hazard_col_name}"
            ] = vlh_trip_type_event
            vlh_total += vlh_trip_type_event
        vehicle_loss_hours.loc[
            [vlh_row.name], f"vlh_{hazard_col_name}_total"
        ] = vlh_total

    def _populate_vehicle_loss_hour(
        self,
        vehicle_loss_hours: gpd.GeoDataFrame,
        row_hazard_range: tuple[float, float],
        vlh_row: pd.Series,
        performance_change: float,
        hazard_col_name: str,
    ):

        vlh_total = 0
        _relevant_link_type = self._get_relevant_link_type(vlh_row, row_hazard_range)
        _divisor = self._get_divisor(_relevant_link_type, row_hazard_range)

        duration_steps = self.resilience_curves.get_duration_steps(
            _relevant_link_type, row_hazard_range
        )
        functionality_loss_ratios = self.resilience_curves.get_functionality_loss_ratio(
            _relevant_link_type, row_hazard_range
        )

        # get vlh_trip_type_event
        for trip_type in self.trip_purposes:
            intensity_trip_type = (
                self.intensities.calculate_intensity(
                    vlh_row[self.link_id], self.traffic_period, trip_type
                )
                / self.hours_per_traffic_period
            )

            vot_trip_type = self.values_of_time.get_value_of_time(trip_type)

            vlh_trip_type_event = sum(
                (
                    intensity_trip_type
                    * duration
                    * loss_ratio
                    * performance_change
                    * vot_trip_type
                )
                / _divisor
                for duration, loss_ratio in zip(
                    duration_steps, functionality_loss_ratios
                )
            )
            vehicle_loss_hours.loc[
                [vlh_row.name], f"vlh_{trip_type}_{hazard_col_name}"
            ] = vlh_trip_type_event
            vlh_total += vlh_trip_type_event
        vehicle_loss_hours.loc[
            [vlh_row.name], f"vlh_{hazard_col_name}_total"
        ] = vlh_total

    @abstractmethod
    def _get_criticality_analysis(self) -> AnalysisLossesProtocol:
        pass

    def execute(self) -> gpd.GeoDataFrame:
        criticality_analysis = self._get_criticality_analysis().execute()

        self._get_disrupted_criticality_analysis_results(
            criticality_analysis=criticality_analysis
        )

        self.result = self.calculate_vehicle_loss_hours()
        return self.result



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

from pyproj import CRS

from ra2ce.common.configuration.config_data_protocol import ConfigDataProtocol
from ra2ce.network.network_config_data.enums.aggregate_wl_enum import AggregateWlEnum
from ra2ce.network.network_config_data.enums.network_type_enum import NetworkTypeEnum
from ra2ce.network.network_config_data.enums.road_type_enum import RoadTypeEnum
from ra2ce.network.network_config_data.enums.source_enum import SourceEnum


@dataclass
class ProjectSection:
    name: str = ""


@dataclass
class NetworkSection:
    directed: bool = False
    source: SourceEnum = field(default_factory=lambda: SourceEnum.INVALID)
    primary_file: list[Path] = field(default_factory=list)
    diversion_file: list[Path] = field(default_factory=list)
    file_id: str = ""
    link_type_column: str = ""
    polygon: Optional[Path] = None
    network_type: NetworkTypeEnum = field(default_factory=lambda: NetworkTypeEnum.NONE)
    road_types: list[RoadTypeEnum] = field(default_factory=list)
    attributes_to_exclude_in_simplification: list[str] = field(default_factory=list)
    save_gpkg: bool = False


@dataclass
class OriginsDestinationsSection:
    origins: Optional[Path] = None
    destinations: Optional[Path] = None
    origins_names: str = ""
    destinations_names: str = ""
    id_name_origin_destination: str = ""
    origin_count: str = ""
    origin_out_fraction: int = (
        1  # fraction of things/people going out of the origin to the destination
    )
    category: str = ""
    region: Optional[Path] = None
    region_var: str = ""


@dataclass
class IsolationSection:
    locations: str = ""  # Should be path.


@dataclass
class HazardSection:
    hazard_map: list[Path] = field(default_factory=list)
    hazard_id: str = ""
    hazard_field_name: str = ""
    aggregate_wl: AggregateWlEnum = field(default_factory=lambda: AggregateWlEnum.NONE)
    hazard_crs: str = ""
    scenario_cost: list[float] = field(default_factory=list)
    # If False no overlay of the segmented network will be created.
    overlay_segmented_network: bool = True


@dataclass
class CleanupSection:
    snapping_threshold: bool = False
    pruning_threshold: bool = False
    segmentation_length: float = float("nan")
    merge_lines: bool = False
    cut_at_intersections: bool = False
    delete_duplicate_nodes: bool = False


@dataclass
class NetworkConfigData(ConfigDataProtocol):
    root_path: Optional[Path] = None
    input_path: Optional[Path] = None
    output_path: Optional[Path] = None
    static_path: Optional[Path] = None
    # CRS is not yet supported in the ini file, it might be relocated to a subsection.
    crs: CRS = field(default_factory=lambda: CRS.from_user_input(4326))
    project: ProjectSection = field(default_factory=ProjectSection)
    network: NetworkSection = field(default_factory=NetworkSection)
    origins_destinations: OriginsDestinationsSection = field(
        default_factory=OriginsDestinationsSection
    )
    isolation: IsolationSection = field(default_factory=IsolationSection)
    hazard: HazardSection = field(default_factory=HazardSection)
    cleanup: CleanupSection = field(default_factory=CleanupSection)

    @property
    def output_graph_dir(self) -> Optional[Path]:
        if not self.static_path:
            return None
        return self.static_path.joinpath("output_graph")

    @property
    def network_dir(self) -> Optional[Path]:
        if not self.static_path:
            return None
        return self.static_path.joinpath("network")

    @staticmethod
    def get_data_output(ini_file: Path) -> Path:
        return ini_file.parent.joinpath("output")

    def to_dict(self) -> dict:
        _dict = self.__dict__
        _dict["crs"] = self.crs.to_epsg()
        _dict["project"] = self.project.__dict__
        _dict["network"] = self.network.__dict__
        _dict["origins_destinations"] = self.origins_destinations.__dict__
        _dict["isolation"] = self.isolation.__dict__
        _dict["hazard"] = self.hazard.__dict__
        _dict["cleanup"] = self.cleanup.__dict__
        return _dict



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007
    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from dataclasses import dataclass

import networkx as nx
from shapely import MultiLineString
from shapely.ops import linemerge
from snkit.network import Network

from ra2ce.network.network_simplification.nx_to_snkit_network_converter import (
    NxToSnkitNetworkConverter,
)
from ra2ce.network.network_simplification.snkit_network_merge_wrapper import merge_edges
from ra2ce.network.network_simplification.snkit_to_nx_network_converter import (
    SnkitToNxNetworkConverter,
)
from ra2ce.network.networks_utils import line_length

NxGraph = nx.Graph | nx.MultiGraph | nx.MultiDiGraph


@dataclass(kw_only=True)
class SnkitNetworkWrapper:
    """
    Wrapper created to reduce complexity of conversion and processing of a `snkit.network.Network`
    within the rest of the code.
    """

    snkit_network: Network
    node_id_column_name: str = "id"
    edge_from_id_column_name: str = "from_id"
    edge_to_id_column_name: str = "to_id"

    @classmethod
    def from_networkx(
        cls,
        networkx_graph: NxGraph,
        column_names_dict: dict[str, str],
    ) -> SnkitNetworkWrapper:
        """
        Generates a `SnkitNetworkWrapper` based on the given `NxGraph`.

        Args:
            networkx_graph (NxGraph): Graph to convert.
            column_names_dict (dict[str, str]): Column names to use.

        Returns:
            SnkitNetworkWrapper: Wrapper containing the converted `snkit.network.Network`.
        """
        _snkit_converted_network = NxToSnkitNetworkConverter(
            networkx_graph=networkx_graph, **column_names_dict
        ).convert()
        return cls(
            snkit_network=_snkit_converted_network,
            **column_names_dict,
        )

    def merge_edges(self, attributes_to_exclude: list[str]) -> None:
        def filter_excluded_attributes() -> list[str]:
            columns_set = set(self.snkit_network.edges.columns)
            return [attr for attr in attributes_to_exclude if attr in columns_set]

        cols = [col for col in self.snkit_network.edges.columns if col != "geometry"]
        _attributes_to_exclude = filter_excluded_attributes()

        if "demand_edge" not in _attributes_to_exclude:
            _aggregate_function = self._aggfunc_with_demand_edge(
                cols, _attributes_to_exclude
            )
        else:
            _aggregate_function = self._aggfunc_no_demand_edge(
                cols, _attributes_to_exclude
            )

        # Overwrite the existing network with the merged edges.
        self.snkit_network = merge_edges(
            snkit_network=self.snkit_network,
            aggregate_func=_aggregate_function,
            by=_attributes_to_exclude,
            id_col="id",
        )

    def process_network(self) -> None:
        _network_crs = self.snkit_network.edges.crs
        self.snkit_network.edges["length"] = self.snkit_network.edges["geometry"].apply(
            lambda x: line_length(x, _network_crs)
        )  # length in m
        self.snkit_network.edges = self.snkit_network.edges[
            self.snkit_network.edges["length"] != 0
        ]  # Remove zero-length edges

        def convert_to_line_string(geometry_to_convert) -> MultiLineString:
            if isinstance(geometry_to_convert, MultiLineString):
                return linemerge([line for line in geometry_to_convert.geoms])
            return geometry_to_convert

        self.snkit_network.edges["geometry"] = self.snkit_network.edges[
            "geometry"
        ].apply(convert_to_line_string)

    def to_networkx(self) -> NxGraph:
        """
        Converts the wrapped `snkit_network` into a corresponding `networkx.Graph`.

        Returns:
            NxGraph: The converted graph.
        """
        return SnkitToNxNetworkConverter(
            snkit_network=self.snkit_network,
            node_id_column_name=self.node_id_column_name,
            edge_from_id_column_name=self.edge_from_id_column_name,
            edge_to_id_column_name=self.edge_to_id_column_name,
        ).convert()

    def _aggfunc_with_demand_edge(self, cols, attributes_to_exclude: list[str]):
        def aggregate_column(col_data, col_name: str):
            if col_name in attributes_to_exclude:
                return col_data.iloc[0]
            elif col_name == "rfid_c":
                return list(col_data)
            elif col_name in ["maxspeed", "avgspeed"]:
                return col_data.mean()
            elif col_name == "demand_edge":
                return max(col_data)
            elif col_data.dtype == "O":
                return "; ".join(
                    str(item) for item in col_data if isinstance(item, str)
                )
            else:
                return col_data.iloc[0]

        return {
            col: (lambda col_data, col_name=col: aggregate_column(col_data, col_name))
            for col in cols
        }

    def _aggfunc_no_demand_edge(self, cols, attributes_to_exclude: list[str]):
        def aggregate_column(col_data, col_name: str):
            if col_name in attributes_to_exclude:
                return col_data.iloc[0]
            elif col_name == "rfid_c":
                return list(col_data)
            elif col_name in ["maxspeed", "avgspeed"]:
                return col_data.mean()
            elif col_data.dtype == "O":
                return "; ".join(
                    str(item) for item in col_data if isinstance(item, str)
                )
            else:
                return col_data.iloc[0]

        return {
            col: (lambda col_data, col_name=col: aggregate_column(col_data, col_name))
            for col in cols
        }



import copy
from pathlib import Path

import geopandas as gpd
import networkx as nx
import numpy as np
import osmnx
import pandas as pd

from ra2ce.analysis.analysis_config_data.analysis_config_data import (
    AnalysisSectionLosses,
)
from ra2ce.analysis.analysis_config_data.enums.weighing_enum import WeighingEnum
from ra2ce.analysis.analysis_input_wrapper import AnalysisInputWrapper
from ra2ce.analysis.losses.analysis_losses_protocol import AnalysisLossesProtocol
from ra2ce.analysis.losses.weighing_analysis.weighing_analysis_factory import (
    WeighingAnalysisFactory,
)
from ra2ce.network.graph_files.graph_file import GraphFile
from ra2ce.network.hazard.hazard_names import HazardNames


class MultiLinkRedundancy(AnalysisLossesProtocol):
    analysis: AnalysisSectionLosses
    graph_file_hazard: GraphFile
    input_path: Path
    static_path: Path
    output_path: Path
    hazard_names: HazardNames

    def __init__(
        self,
        analysis_input: AnalysisInputWrapper,
    ) -> None:
        self.analysis = analysis_input.analysis
        self.graph_file_hazard = analysis_input.graph_file_hazard
        self.input_path = analysis_input.input_path
        self.static_path = analysis_input.static_path
        self.output_path = analysis_input.output_path
        self.hazard_names = analysis_input.hazard_names

    def _update_time(
        self, df_calculated: pd.DataFrame, gdf_graph: gpd.GeoDataFrame
    ) -> tuple[pd.DataFrame, gpd.GeoDataFrame]:
        """
        updates the time column with the calculated dataframe and updates the rest of the gdf_graph if time is None.
        """
        if WeighingEnum.TIME.config_value not in df_calculated.columns:
            return df_calculated, gdf_graph

        if (
            WeighingEnum.TIME.config_value in gdf_graph.columns
            and WeighingEnum.TIME.config_value in df_calculated.columns
        ):
            df_calculated = df_calculated.drop(columns=[WeighingEnum.TIME.config_value])
            return df_calculated, gdf_graph

        gdf_graph[WeighingEnum.TIME.config_value] = df_calculated[
            WeighingEnum.TIME.config_value
        ]
        for i, row in gdf_graph.iterrows():
            row_avgspeed = row.get("avgspeed", None)
            row_length = row.get("length", None)
            if (
                pd.isna(row[WeighingEnum.TIME.config_value])
                and row_avgspeed
                and row_length
            ):
                gdf_graph.at[i, WeighingEnum.TIME.config_value] = (
                    row_length * 1e-3 / row_avgspeed
                )
            else:
                gdf_graph.at[i, WeighingEnum.TIME.config_value] = row.get(
                    WeighingEnum.TIME.config_value, None
                )
        return df_calculated, gdf_graph

    def execute(self) -> gpd.GeoDataFrame:
        """Calculates the multi-link redundancy of a NetworkX graph.

        The function removes all links of a variable that have a minimum value
        of min_threshold. For each link it calculates the alternative path, if
        any available. This function only removes one group at the time and saves the data from removing that group.

        Returns:
            aggregated_results (GeoDataFrame): The results of the analysis aggregated into a table.
        """

        def _is_not_none(value):
            return (
                value is not None
                and value is not pd.NA
                and not pd.isna(value)
                and not np.isnan(value)
            )

        results = []
        master_graph = copy.deepcopy(self.graph_file_hazard.get_graph())
        for hazard in self.hazard_names.names:
            hazard_name = self.hazard_names.get_name(hazard)

            _graph = copy.deepcopy(master_graph)
            # Create a geodataframe from the full graph
            gdf = osmnx.graph_to_gdfs(master_graph, nodes=False)
            if "rfid" in gdf:
                gdf["rfid"] = gdf["rfid"].astype(str)

            # Create the edgelist that consist of edges that should be removed
            edges_remove = []
            for e in _graph.edges.data(keys=True):
                if (hazard_name in e[-1]) and (
                    ("bridge" not in e[-1])
                    or ("bridge" in e[-1] and e[-1]["bridge"] != "yes")
                ):
                    edges_remove.append(e)
            edges_remove = [e for e in edges_remove if (e[-1][hazard_name] is not None)]
            edges_remove = [
                e
                for e in edges_remove
                if (hazard_name in e[-1])
                and (
                    _is_not_none(e[-1][hazard_name])
                    and (e[-1][hazard_name] > float(self.analysis.threshold))
                    and (
                        ("bridge" not in e[-1])
                        or ("bridge" in e[-1] and e[-1]["bridge"] != "yes")
                    )
                )
            ]

            _graph.remove_edges_from(edges_remove)

            columns = [
                "u",
                "v",
                f"alt_{self.analysis.weighing.config_value}",
                "alt_nodes",
                f"diff_{self.analysis.weighing.config_value}",
                "connected",
            ]

            if "rfid" in gdf:
                columns.insert(2, "rfid")

            df_calculated = pd.DataFrame(columns=columns)
            _weighing_analyser = WeighingAnalysisFactory.get_analysis(
                self.analysis.weighing
            )

            for edges in edges_remove:
                u, v, _, _weighing_analyser.edge_data = edges
                _current_value = _weighing_analyser.get_current_value()

                _alt_value, _alt_nodes, _connected, _diff = np.nan, np.nan, 0, np.nan
                if nx.has_path(_graph, u, v):
                    [_alt_value, _alt_nodes] = nx.single_source_dijkstra(
                        _graph,
                        u,
                        v,
                        weight=self.analysis.weighing.config_value,
                    )
                    _connected = 1

                    _diff = round(_alt_value - _current_value, 3)

                data = {
                    "u": u,
                    "v": v,
                    self.analysis.weighing.config_value: _current_value,
                    f"alt_{self.analysis.weighing.config_value}": _alt_value,
                    "alt_nodes": [_alt_nodes],
                    f"diff_{self.analysis.weighing.config_value}": _diff,
                    "connected": _connected,
                }

                if "rfid" in gdf:
                    data["rfid"] = [str(_weighing_analyser.edge_data["rfid"])]

                df_calculated = pd.concat(
                    [df_calculated, pd.DataFrame(data)], ignore_index=True
                )

            df_calculated[f"alt_{self.analysis.weighing.config_value}"] = pd.to_numeric(
                df_calculated[f"alt_{self.analysis.weighing.config_value}"],
                errors="coerce",
            )

            df_calculated, gdf = self._update_time(df_calculated, gdf)

            # Merge the dataframes
            if "rfid" in gdf:
                gdf = gdf.merge(df_calculated, how="left", on=["u", "v", "rfid"])
            else:
                gdf = gdf.merge(df_calculated, how="left", on=["u", "v"])

            gdf["hazard"] = hazard_name

            results.append(gdf)

        return pd.concat(results, ignore_index=True)



from typing import Any, Protocol, runtime_checkable

import geopandas as gpd

from ra2ce.analysis.analysis_config_data.enums.damage_curve_enum import DamageCurveEnum


@runtime_checkable
class ToIntegrateShaperProtocol(Protocol):
    gdf: gpd.GeoDataFrame

    def get_return_periods(self) -> list:
        """
        Gets the columns' names that have damages calculated

        Returns:
            list: columns' name that have damages calculated.
        """

    def shape_to_integrate_object(
        self, return_periods: list
    ) -> dict[str : gpd.GeoDataFrame]:
        """
        Gets the return periods and create columns for risk calculation for each damage curve and return period
        Returns:
            list[GeoDataFrame]:
            to_integrate objects each corresponding to each vulnerability curve to be used for risk calculation
        """



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import math

import geopandas as gpd
import networkx as nx
import pandas as pd
from shapely.geometry import MultiLineString

import ra2ce.network.networks_utils as nut
from ra2ce.network.network_config_data.network_config_data import NetworkConfigData
from ra2ce.network.network_wrappers.network_wrapper_protocol import (
    NetworkWrapperProtocol,
)
from ra2ce.network.segmentation import Segmentation


class ShpNetworkWrapper(NetworkWrapperProtocol):
    def __init__(
        self,
        config_data: NetworkConfigData,
    ) -> None:
        _network_options = config_data.network
        _cleanup_options = config_data.cleanup

        self.project_name = config_data.project.name
        self.output_graph_dir = config_data.output_graph_dir
        self.crs = config_data.crs

        # Network options
        self.primary_files = _network_options.primary_file
        self.diversion_files = _network_options.diversion_file
        self.directed = _network_options.directed
        self.file_id = _network_options.file_id

        # Cleanup options
        self.merge_lines = _cleanup_options.merge_lines
        self.snapping_threshold = _cleanup_options.snapping_threshold
        self.segmentation_length = _cleanup_options.segmentation_length
        self.cut_at_intersections = _cleanup_options.cut_at_intersections

        # Origins Destinations
        self.region_path = config_data.origins_destinations.region

    def _read_merge_shp(self) -> gpd.GeoDataFrame:
        """Imports shapefile(s) and saves attributes in a pandas dataframe.

        Returns:
            lines (list of shapely LineStrings): full list of linestrings
            properties (pandas dataframe): attributes of shapefile(s), in order of the linestrings in lines
        """
        # concatenate all shapefile into one geodataframe and set analysis to 1 or 0 for diversions
        lines = [gpd.read_file(shp, engine="pyogrio") for shp in self.primary_files]

        if any(self.diversion_files):
            lines.extend(
                [
                    nut.check_crs_gdf(gpd.read_file(shp, engine="pyogrio"), self.crs)
                    for shp in self.diversion_files
                ]
            )
        lines = pd.concat(lines)

        lines.crs = self.crs

        # Check if there are any multilinestrings and convert them to linestrings.
        if lines["geometry"].apply(lambda row: isinstance(row, MultiLineString)).any():
            mls_idx = lines.loc[
                lines["geometry"].apply(lambda row: isinstance(row, MultiLineString))
            ].index
            for idx in mls_idx:
                # Multilinestrings to linestrings
                new_rows_geoms = list(lines.iloc[idx]["geometry"].geoms)
                for nrg in new_rows_geoms:
                    dict_attributes = dict(lines.iloc[idx])
                    dict_attributes["geometry"] = nrg
                    lines.loc[max(lines.index) + 1] = dict_attributes

            lines = lines.drop(labels=mls_idx, axis=0)

        # append the length of the road stretches
        lines["length"] = lines["geometry"].apply(
            lambda x: nut.line_length(x, self.crs)
        )

        logging.info(
            "Shapefile(s) loaded with attributes: {}.".format(
                list(lines.columns.values)
            )
        )  # fill in parameter names

        return lines

    def _get_complex_graph_and_edges(
        self, edges: gpd.GeoDataFrame, id_name: str
    ) -> tuple[nx.MultiGraph, gpd.GeoDataFrame]:
        # Get the unique points at the end of lines and at intersections to create nodes
        nodes = nut.create_nodes(edges, self.crs, self.cut_at_intersections)
        logging.info("Function [create_nodes]: executed")

        edges = nut.cut_lines(
            edges, nodes, id_name, tolerance=0.00001, crs_=self.crs
        )  ## PAY ATTENTION TO THE TOLERANCE, THE UNIT IS DEGREES
        logging.info("Function [cut_lines]: executed")

        if not edges.crs:
            edges.crs = self.crs

        # create tuples from the adjacent nodes and add as column in geodataframe
        edges_complex = nut.join_nodes_edges(nodes, edges, id_name)
        edges_complex.crs = self.crs  # set the right CRS
        edges_complex.dropna(subset=["node_A", "node_B"], inplace=True)

        assert (
            edges_complex["node_A"].isnull().sum() == 0
        ), "Some edges cannot be assigned nodes, please check your input shapefile."
        assert (
            edges_complex["node_B"].isnull().sum() == 0
        ), "Some edges cannot be assigned nodes, please check your input shapefile."

        # Create networkx graph from geodataframe
        graph_complex = nut.graph_from_gdf(edges_complex, nodes, node_id="node_fid")
        logging.info("Function [graph_from_gdf]: executed")
        return graph_complex, edges_complex

    def get_network(
        self,
    ) -> tuple[nx.MultiGraph, gpd.GeoDataFrame]:
        edges = self._read_merge_shp()
        lines_merged = gpd.GeoDataFrame()
        # Check which of the lines are merged, also for the fid. The fid of the first line with a traffic count is taken.
        # The list of fid's is reduced by the fid's that are not anymore in the merged lines
        if self.merge_lines:
            aadt_names = []
            edges, lines_merged = nut.merge_lines_automatic(
                edges, self.file_id, aadt_names, self.crs
            )
            logging.info(
                "Function [merge_lines_automatic]: executed with properties {}".format(
                    list(edges.columns)
                )
            )

        edges, id_name = nut.gdf_check_create_unique_ids(edges, self.file_id)

        if self.snapping_threshold:
            # TODO: snapping threshold it's a bool yet here we expect a float.
            edges = nut.snap_endpoints_lines(
                edges, self.snapping_threshold, id_name, self.crs
            )
            logging.info(
                "Function [snap_endpoints_lines]: executed with threshold = {}".format(
                    self.snapping_threshold
                )
            )

        # merge merged lines if there are any merged lines
        if not lines_merged.empty:
            # save the merged lines to a shapefile - CHECK if there are lines merged that should not be merged (e.g. main + secondary road)
            lines_merged.set_geometry(
                col="geometry", inplace=True
            )  # To ensure the object is a GeoDataFrame and not a Series
            _emerged_lines_file = self.output_graph_dir.joinpath(
                f"{self.project_name}_lines_that_merged.gpkg"
            )
            lines_merged.to_file(_emerged_lines_file)
            logging.info(
                "Function [edges_to_shp]: saved at {}".format(_emerged_lines_file)
            )

        graph_complex, edges_complex = self._get_complex_graph_and_edges(edges, id_name)

        if not math.isnan(self.segmentation_length):
            edges_complex = Segmentation(edges_complex, self.segmentation_length)
            edges_complex = edges_complex.apply_segmentation()
            if edges_complex.crs is None:  # The CRS might have dissapeared.
                edges_complex.crs = self.crs  # set the right CRS
        return graph_complex, edges_complex



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import pickle
from pathlib import Path
from typing import Optional

from ra2ce.network.exporters.network_exporter_base import (
    MULTIGRAPH_TYPE,
    NetworkExporterBase,
)
from ra2ce.network.networks_utils import graph_to_gpkg


class MultiGraphNetworkExporter(NetworkExporterBase):
    pickle_path: Optional[Path]

    def export_to_gpkg(self, output_dir: Path, export_data: MULTIGRAPH_TYPE) -> None:
        if not output_dir.is_dir():
            output_dir.mkdir(parents=True)

        # TODO: This method should be a writer itself.
        graph_to_gpkg(
            export_data,
            output_dir / (self._basename + "_edges.gpkg"),
            output_dir / (self._basename + "_nodes.gpkg"),
        )
        logging.info(
            f"Saved {self._basename + '_edges.gpkg'} and {self._basename + '_nodes.gpkg'} in {output_dir}."
        )

    def export_to_pickle(self, output_dir: Path, export_data: MULTIGRAPH_TYPE) -> None:
        self.pickle_path = output_dir / (self._basename + ".p")
        with open(self.pickle_path, "wb") as f:
            pickle.dump(export_data, f, protocol=4)
        logging.info(
            f"Saved {self.pickle_path.stem} in {self.pickle_path.resolve().parent}."
        )



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


from pathlib import Path

import geopandas as gpd
import networkx as nx

from ra2ce.common.io.writers.ra2ce_exporter_protocol import Ra2ceExporterProtocol

MULTIGRAPH_TYPE = nx.MultiGraph | nx.MultiDiGraph
NETWORK_TYPE = gpd.GeoDataFrame | MULTIGRAPH_TYPE


class NetworkExporterBase(Ra2ceExporterProtocol):
    _basename: str
    _export_types: list[str] = ["pickle"]
    pickle_path: Path

    def __init__(self, basename: str, export_types: list[str]) -> None:
        self._basename = basename
        self._export_types = export_types
        self.pickle_path = None

    def export_to_gpkg(self, output_dir: Path, export_data: NETWORK_TYPE) -> None:
        """
        Exports the given data into a `*.gpkg` file.

        Args:
            output_dir (Path): Output directory where the save the exported data.
            export_data (NETWORK_TYPE): Data that needs to be exported.
        """
        pass

    def export_to_pickle(self, output_dir: Path, export_data: NETWORK_TYPE) -> None:
        """
        Exports the given data into a `*.p` or `.feather` file.

        Args:
            output_dir (Path): Output directory where the save the exported data.
            export_data (NETWORK_TYPE): Data that needs to be exported.
        """
        pass

    def export(self, export_path: Path, export_data: NETWORK_TYPE) -> None:
        """
        Exports the given data to the specified types.
        TODO: I do not particularly like this approach, but at least gives better clarity on what the export options at the moment are.

        Args:
            export_path (Path): Path to the output directory where to export the data.
            export_data (NETWORK_TYPE): Data that needs to be exported.
        """
        if "pickle" in self._export_types:
            self.export_to_pickle(export_path, export_data)

        if "gpkg" in self._export_types:
            self.export_to_gpkg(export_path, export_data)



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from ra2ce.analysis.analysis_config_data.analysis_config_data import (
    AnalysisSectionDamages,
    AnalysisSectionLosses,
)
from ra2ce.analysis.analysis_config_data.enums.analysis_damages_enum import (
    AnalysisDamagesEnum,
)
from ra2ce.analysis.analysis_config_data.enums.analysis_losses_enum import (
    AnalysisLossesEnum,
)
from ra2ce.analysis.analysis_config_wrapper import AnalysisConfigWrapper
from ra2ce.analysis.analysis_input_wrapper import AnalysisInputWrapper
from ra2ce.analysis.damages.analysis_damages_protocol import AnalysisDamagesProtocol
from ra2ce.analysis.damages.damages import Damages
from ra2ce.analysis.damages.effectiveness_measures import EffectivenessMeasures
from ra2ce.analysis.losses.analysis_losses_protocol import AnalysisLossesProtocol
from ra2ce.analysis.losses.multi_link_isolated_locations import (
    MultiLinkIsolatedLocations,
)
from ra2ce.analysis.losses.multi_link_losses import MultiLinkLosses
from ra2ce.analysis.losses.multi_link_origin_closest_destination import (
    MultiLinkOriginClosestDestination,
)
from ra2ce.analysis.losses.multi_link_origin_destination import (
    MultiLinkOriginDestination,
)
from ra2ce.analysis.losses.multi_link_redundancy import MultiLinkRedundancy
from ra2ce.analysis.losses.optimal_route_origin_closest_destination import (
    OptimalRouteOriginClosestDestination,
)
from ra2ce.analysis.losses.optimal_route_origin_destination import (
    OptimalRouteOriginDestination,
)
from ra2ce.analysis.losses.single_link_losses import SingleLinkLosses
from ra2ce.analysis.losses.single_link_redundancy import SingleLinkRedundancy


class AnalysisFactory:
    @staticmethod
    def get_damages_analysis(
        analysis: AnalysisSectionDamages,
        analysis_config: AnalysisConfigWrapper,
    ) -> AnalysisDamagesProtocol:
        """
        Create an analysis based on the given analysis configuration.

        Args:
            analysis (AnalysisSectionDamages): Analysis section.
            analysis_config (AnalysisConfigWrapper): Analysis configuration.

        Raises:
            NotImplementedError: The analysis type is not implemented.

        Returns:
            AnalysisDamagesProtocol: The damages analysis to be executed.
        """
        _analysis_input = AnalysisInputWrapper.from_input(
            analysis=analysis,
            analysis_config=analysis_config,
            graph_file_hazard=analysis_config.graph_files.base_network_hazard,
        )

        if analysis.analysis == AnalysisDamagesEnum.DAMAGES:
            return Damages(_analysis_input)

        if analysis.analysis == AnalysisDamagesEnum.EFFECTIVENESS_MEASURES:
            return EffectivenessMeasures(_analysis_input)

        raise NotImplementedError(f"Analysis {analysis.analysis} not implemented")

    @staticmethod
    def get_losses_analysis(
        analysis: AnalysisSectionLosses,
        analysis_config: AnalysisConfigWrapper,
    ) -> AnalysisLossesProtocol:
        """
        Create an analysis based on the given analysis configuration.

        Args:
            analysis (AnalysisSectionLosses): Analysis section.
            analysis_config (AnalysisConfigWrapper): Analysis configuration.

        Raises:
            NotImplementedError: The analysis type is not implemented.

        Returns:
            AnalysisLossesProtocol: The losses analysis to be executed.
        """
        if analysis.analysis == AnalysisLossesEnum.SINGLE_LINK_REDUNDANCY:
            _analysis_input = AnalysisInputWrapper.from_input(
                analysis=analysis,
                analysis_config=analysis_config,
                graph_file=analysis_config.graph_files.base_graph,
            )
            return SingleLinkRedundancy(_analysis_input)

        if analysis.analysis == AnalysisLossesEnum.MULTI_LINK_REDUNDANCY:
            _analysis_input = AnalysisInputWrapper.from_input(
                analysis=analysis,
                analysis_config=analysis_config,
                graph_file_hazard=analysis_config.graph_files.base_graph_hazard,
            )
            return MultiLinkRedundancy(_analysis_input)

        if analysis.analysis == AnalysisLossesEnum.OPTIMAL_ROUTE_ORIGIN_DESTINATION:
            _analysis_input = AnalysisInputWrapper.from_input(
                analysis=analysis,
                analysis_config=analysis_config,
                graph_file=analysis_config.graph_files.origins_destinations_graph,
            )
            return OptimalRouteOriginDestination(_analysis_input)

        if analysis.analysis == AnalysisLossesEnum.MULTI_LINK_ORIGIN_DESTINATION:
            _analysis_input = AnalysisInputWrapper.from_input(
                analysis=analysis,
                analysis_config=analysis_config,
                graph_file_hazard=analysis_config.graph_files.origins_destinations_graph_hazard,
            )
            return MultiLinkOriginDestination(_analysis_input)

        if (
            analysis.analysis
            == AnalysisLossesEnum.OPTIMAL_ROUTE_ORIGIN_CLOSEST_DESTINATION
        ):
            _analysis_input = AnalysisInputWrapper.from_input(
                analysis=analysis,
                analysis_config=analysis_config,
                graph_file_hazard=analysis_config.graph_files.origins_destinations_graph_hazard,
            )
            return OptimalRouteOriginClosestDestination(analysis_input=_analysis_input)

        if (
            analysis.analysis
            == AnalysisLossesEnum.MULTI_LINK_ORIGIN_CLOSEST_DESTINATION
        ):
            _analysis_input = AnalysisInputWrapper.from_input(
                analysis=analysis,
                analysis_config=analysis_config,
                graph_file=analysis_config.graph_files.origins_destinations_graph,
                graph_file_hazard=analysis_config.graph_files.origins_destinations_graph_hazard,
            )
            return MultiLinkOriginClosestDestination(_analysis_input)

        if analysis.analysis == AnalysisLossesEnum.SINGLE_LINK_LOSSES:
            _analysis_input = AnalysisInputWrapper.from_input(
                analysis=analysis,
                analysis_config=analysis_config,
                graph_file=analysis_config.graph_files.base_graph_hazard,
                graph_file_hazard=analysis_config.graph_files.base_graph_hazard,
            )
            return SingleLinkLosses(_analysis_input, analysis_config)

        if analysis.analysis == AnalysisLossesEnum.MULTI_LINK_LOSSES:
            _analysis_input = AnalysisInputWrapper.from_input(
                analysis=analysis,
                analysis_config=analysis_config,
                graph_file_hazard=analysis_config.graph_files.base_graph_hazard,
            )
            return MultiLinkLosses(_analysis_input, analysis_config)

        if analysis.analysis == AnalysisLossesEnum.MULTI_LINK_ISOLATED_LOCATIONS:
            _analysis_input = AnalysisInputWrapper.from_input(
                analysis=analysis,
                analysis_config=analysis_config,
                graph_file_hazard=analysis_config.graph_files.base_graph_hazard,
            )
            return MultiLinkIsolatedLocations(_analysis_input)

        raise NotImplementedError(f"Analysis {analysis.analysis} not implemented")



import geopandas as gpd

from ra2ce.analysis.analysis_config_data.enums.damage_curve_enum import DamageCurveEnum
from ra2ce.analysis.damages.shape_to_integrate_object.to_Integrate_shaper_protocol import (
    ToIntegrateShaperProtocol,
)


class HzToIntegrateShaper(ToIntegrateShaperProtocol):
    gdf: gpd.GeoDataFrame

    def __init__(self, gdf):
        self.gdf = gdf

    def get_return_periods(self) -> list:
        return sorted(c for c in self.gdf.columns if c.startswith("dam"))

    def shape_to_integrate_object(
        self, return_periods: list
    ) -> dict[str : gpd.GeoDataFrame]:
        _to_integrate = self.gdf[return_periods]

        _to_integrate.columns = [
            float(c.split("_")[1].replace("RP", "")) for c in _to_integrate.columns
        ]
        return {
            DamageCurveEnum.HZ.name: _to_integrate.sort_index(
                axis="columns", ascending=False
            )
        }



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007
    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from dataclasses import dataclass

import networkx as nx

from ra2ce.network.network_simplification.snkit_network_wrapper import (
    SnkitNetworkWrapper,
)

NxGraph = nx.Graph | nx.MultiGraph | nx.MultiDiGraph


@dataclass(kw_only=True)
class NetworkSimplificationWithAttributeExclusion:
    """
    Simplifies a network by excluding a given set of attributes (columns).
    """

    nx_graph: NxGraph
    attributes_to_exclude: list[str]

    def simplify_graph(self) -> nx.Graph:
        """
        Simplifies the inner graph by using the `snkit` package.

        Returns:
            nx.Graph: Resulting simplified graph.
        """
        _snkit_network_wrapper = SnkitNetworkWrapper.from_networkx(
            self.nx_graph,
            column_names_dict=dict(
                node_id_column_name="id",
                edge_from_id_column_name="from_id",
                edge_to_id_column_name="to_id",
            ),
        )
        _snkit_network_wrapper.merge_edges(self.attributes_to_exclude)
        _snkit_network_wrapper.process_network()
        return _snkit_network_wrapper.to_networkx()



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007
    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import re
from pathlib import Path

import pandas as pd

from ra2ce.analysis.analysis_config_data.enums.traffic_period_enum import (
    TrafficPeriodEnum,
)
from ra2ce.analysis.analysis_config_data.enums.trip_purpose_enum import TripPurposeEnum
from ra2ce.analysis.losses.losses_input_data_reader_base import (
    LossesInputDataReaderBase,
)
from ra2ce.analysis.losses.traffic_intensities.traffic_intensities import (
    TrafficIntensities,
)


class TrafficIntensitiesReader(LossesInputDataReaderBase):
    """
    Class to read the traffic intensities per traffic period from a csv file.
    """

    csv_columns = []
    separator = ","
    data_type = TrafficIntensities

    def __init__(self, link_id: str) -> None:
        self.csv_columns = [link_id]

    def _parse_df(self, df: pd.DataFrame) -> TrafficIntensities:
        _traffic_intensities = TrafficIntensities()
        for col in df:
            if col == self.csv_columns[0]:
                _traffic_intensities.link_id = df[col].tolist()
                continue
            _col_parts = re.findall(r"(.+)_(\w+)", col)  # split on last underscore
            _traffic_period = TrafficPeriodEnum.get_enum(_col_parts[0][0])
            _trip_purpose = TripPurposeEnum.get_enum(_col_parts[0][1])
            _traffic_intensities.intensities[(_traffic_period, _trip_purpose)] = df[
                col
            ].tolist()
        return _traffic_intensities

    def read(self, file_path: Path | None) -> TrafficIntensities:
        return super().read(file_path)



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import time

from ra2ce.analysis.analysis_collection import AnalysisCollection
from ra2ce.analysis.analysis_config_wrapper import AnalysisConfigWrapper
from ra2ce.analysis.analysis_result_wrapper import AnalysisResultWrapper
from ra2ce.analysis.analysis_result_wrapper_exporter import (
    AnalysisResultWrapperExporter,
)
from ra2ce.configuration.config_wrapper import ConfigWrapper
from ra2ce.runners.analysis_runner_protocol import AnalysisRunner


class LossesAnalysisRunner(AnalysisRunner):
    def __str__(self) -> str:
        return "Losses Analysis Runner"

    @staticmethod
    def can_run(ra2ce_input: ConfigWrapper) -> bool:
        if (
            not ra2ce_input.analysis_config
            or not ra2ce_input.analysis_config.config_data.losses_list
        ):
            return False
        return True

    def run(
        self, analysis_config: AnalysisConfigWrapper
    ) -> list[AnalysisResultWrapper]:
        _analysis_collection = AnalysisCollection.from_config(analysis_config)
        _results = []
        for analysis in _analysis_collection.losses_analyses:
            logging.info(
                "----------------------------- Started analyzing '%s'  -----------------------------",
                analysis.analysis.name,
            )
            starttime = time.time()

            _result = analysis.execute()
            _result_wrapper = AnalysisResultWrapper(
                analysis_result=_result, analysis=analysis
            )

            _results.append(_result_wrapper)
            AnalysisResultWrapperExporter().export_result(_result_wrapper)

            endtime = time.time()
            logging.info(
                "----------------------------- Analysis '%s' finished. "
                "Time: %ss  -----------------------------",
                analysis.analysis.name,
                str(round(endtime - starttime, 2)),
            )
        return _results



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007
    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from dataclasses import dataclass, field

from ra2ce.analysis.analysis_config_data.enums.trip_purpose_enum import TripPurposeEnum


@dataclass(kw_only=True)
class TimeValues:
    """
    Class to store the time values for different trip types.
    """

    time_values: dict[TripPurposeEnum, tuple[int, int]] = field(default_factory=dict)

    def get_value_of_time(self, trip_type: TripPurposeEnum) -> int:
        """
        Get the value of time for a given trip type.

        Args:
            trip_type (TripPurposeEnum): The type of the trip.

        Returns:
            int: The value of time.
        """
        return self.time_values[trip_type][0]

    def get_occupants(self, trip_type: TripPurposeEnum) -> int:
        """
        Get the occupants for a given trip type.

        Args:
            trip_type (TripPurposeEnum): The type of the trip.

        Returns:
            int: The occupants.
        """
        return self.time_values[trip_type][1]



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import os
from pathlib import Path
from typing import Optional, Union

import geopandas as gpd
import networkx as nx
import numpy as np
import pandas as pd
import pyproj
import rasterio
import rasterio.mask
import rasterio.transform
from rasterio import Affine
from rasterio.warp import Resampling, calculate_default_transform, reproject
from shapely.geometry import LineString, Point
from tqdm import tqdm

from ra2ce.network.networks_utils import cut, line_length

"""
TODO: This whole file should be throughouly tested / redesigned.
"""


def read_origin_destination_files(
    origin_paths: Union[str, list],
    origin_names: Union[str, list],
    destination_paths: Union[str, list],
    destination_names: Union[str, list],
    od_id: str,
    origin_count: str,
    crs_: pyproj.CRS,
    category: str,
    region_paths: Optional[str],
    region_var: Optional[str],
):
    """Reads the Origin and Destination point shapefiles and creates one big OD GeoDataFrame.
    Args:
        origin_paths: The path (as string) or paths (in a list) of the point shapefile(s) used for the locations of the Origins.
        origin_names: The name(s) of the origins
        destination_paths: The path (as string) or paths (in a list) of the point shapefile(s) used for the locations of the Destinations.
        destination_names: The name(s) of the destinations
        od_id: The name of the unique identifier attribute in both the origin and destination shapefiles.
        origin_count: The name of the attribute in the origin shapefile that can be used for counting the flow over the network (e.g. nr. of people)
        crs_: The Coordinate Reference System used in the project.
        category: The name of the attribute in the destination shapefile that can be used to categorize the (closest) destination.
        region_paths:
        region_var:
    Returns:
        od:
    """

    if region_paths:
        origin = gpd.GeoDataFrame(
            columns=[od_id, "o_id", "geometry", "region"], crs=crs_
        )
        region = gpd.read_file(region_paths, engine="pyogrio")
        region = region[[region_var, "geometry"]]
    else:
        origin = gpd.GeoDataFrame(columns=[od_id, "o_id", "geometry"], crs=crs_)

    destination_columns = [od_id, "d_id", "geometry"]
    if category:
        destination_columns.append(category)

    destination = gpd.GeoDataFrame(columns=destination_columns, crs=crs_)

    if isinstance(origin_paths, str):
        origin_paths = [origin_paths]
    if isinstance(destination_paths, str):
        destination_paths = [destination_paths]
    if isinstance(origin_names, str):
        origin_names = [origin_names]
    if isinstance(destination_names, str):
        destination_names = [destination_names]

    for op, on in zip(origin_paths, origin_names):
        origin_new = gpd.read_file(op, crs=crs_, engine="pyogrio")
        try:
            origin_new[od_id] * 2  # just for checking
        except Exception:
            origin_new[od_id] = origin_new.index

        if region_paths:
            origin_new = gpd.sjoin(left_df=origin_new, right_df=region, how="left")
            origin_new["region"] = origin_new[region_var]
            origin_new = origin_new[[od_id, origin_count, "geometry", "region"]]
            origin_new["region"].fillna("Not assigned", inplace=True)
        else:
            origin_new = origin_new[[od_id, origin_count, "geometry"]]
        origin_new["o_id"] = on + "_" + origin_new[od_id].astype(str)
        origin_new.crs = origin.crs
        origin = gpd.GeoDataFrame(pd.concat([origin, origin_new], ignore_index=True))

    destination_columns_add = [od_id, "geometry"]
    if category:
        destination_columns_add.append(category)

    for dp, dn in zip(destination_paths, destination_names):
        destination_new = gpd.read_file(dp, crs=crs_, engine="pyogrio")
        try:
            assert destination_new[od_id]
        except Exception:
            destination_new[od_id] = destination_new.index
        destination_new = destination_new[destination_columns_add]
        destination_new["d_id"] = dn + "_" + destination_new[od_id].astype(str)
        destination_new.crs = destination.crs
        destination = gpd.GeoDataFrame(
            pd.concat([destination, destination_new], ignore_index=True)
        )

    od = pd.concat([origin, destination], sort=False)

    return od


def closest_node(node: np.ndarray, nodes: np.ndarray) -> np.ndarray:
    deltas = nodes - node
    dist_2 = np.einsum("ij,ij->i", deltas, deltas)
    # Find indices of all minimum distances
    min_dist_indices = np.where(dist_2 == dist_2.min())[0]

    # CROSSCADE: The previous return value `nodes[np.argmin(dist_2)]`
    # returned a np.ndarray of 1 dimension (`[result]`), whilst the new value
    # will return a 2 dimensional np.ndarray ( `[[result]]`)
    return np.round(nodes[min_dist_indices], decimals=7)[0]


def get_od(o_id: str, d_id: str) -> str:
    """
    Gets a valid origin id node from the given pair.

    Args:
        o_id (str): Id for the `origin` node.
        d_id (str): Id for the `destination` node.

    Returns:
        str | np.nan: Valid value to represent the origin - destination node.
    """
    _nan_values = ["nan", np.nan]
    if o_id not in _nan_values:
        return o_id
    if d_id not in _nan_values:
        # `o_id` was nan, so it was a destination, not an origin.
        # therefore we return `d_id`
        return d_id
    return np.nan


def add_data_to_existing_node(graph, node, match_name):
    if "od_id" in graph.nodes[node]:
        # the node already has a origin/destination attribute
        if match_name not in set(graph.nodes[node]["od_id"].split(",")):
            graph.nodes[node]["od_id"] = graph.nodes[node]["od_id"] + "," + match_name
    else:
        graph.nodes[node]["od_id"] = match_name
    return graph


def update_edges_with_new_node(
    graph: nx.MultiGraph,
    edge_data: dict,
    node_a: int,
    node_b: int,
    k: int,
    line_a: LineString,
    line_b: LineString,
    new_node_id: int,
    graph_crs: pyproj.CRS,
    inverse_vertices_dict: dict,
):
    # Check which line is connected to which node. There can be 8 different combinations and there should be two
    # edges added to the graph.
    cnt = 0

    if Point(graph.nodes[node_a]["geometry"].coords[0]).almost_equals(
        Point(line_b.coords[-1])
    ):
        if node_a == node_b and graph.has_edge(*(node_a, new_node_id, 0)):
            if line_b != graph.edges[(node_a, new_node_id, 0)]["geometry"]:
                k_new = 1
            else:
                k_new = 0
        else:
            k_new = 0
        edge_data.update(
            length=line_length(line_b, graph_crs),
            geometry=line_b,
            node_A=node_a,
            node_B=new_node_id,
            edge_fid=f"{node_a}_{new_node_id}",
        )
        graph.add_edge(node_a, new_node_id, k_new, **edge_data)

        # Update the inverse vertices dict
        inverse_vertices_dict.update(
            {p: (node_a, new_node_id, k_new) for p in set(list(line_b.coords[1:-1]))}
        )

        cnt += 1

    if Point(graph.nodes[node_b]["geometry"].coords[0]).almost_equals(
        Point(line_b.coords[0])
    ):
        if node_a == node_b and graph.has_edge(*(node_a, new_node_id, 0)):
            if line_b != graph.edges[(node_a, new_node_id, 0)]["geometry"]:
                k_new = 1
            else:
                k_new = 0
        else:
            k_new = 0
        edge_data.update(
            length=line_length(line_b, graph_crs),
            geometry=line_b,
            node_A=new_node_id,
            node_B=node_b,
            edge_fid=f"{new_node_id}_{node_b}",
        )
        graph.add_edge(
            new_node_id, node_b, k_new, **edge_data
        )  # changed the u, v order

        # Update the inverse vertices dict
        inverse_vertices_dict.update(
            {p: (node_b, new_node_id, k_new) for p in set(list(line_b.coords[1:-1]))}
        )

        cnt += 1

    if Point(graph.nodes[node_a]["geometry"].coords[0]).almost_equals(
        Point(line_b.coords[0])
    ):
        if node_a == node_b and graph.has_edge(*(node_a, new_node_id, 0)):
            if line_b != graph.edges[(node_a, new_node_id, 0)]["geometry"]:
                k_new = 1
            else:
                k_new = 0
        else:
            k_new = 0
        edge_data.update(
            length=line_length(line_b, graph_crs),
            geometry=line_b,
            node_A=node_a,
            node_B=new_node_id,
            edge_fid=f"{node_b}_{new_node_id}",
        )
        graph.add_edge(node_a, new_node_id, k_new, **edge_data)

        # Update the inverse vertices dict
        inverse_vertices_dict.update(
            {p: (node_a, new_node_id, k_new) for p in set(list(line_b.coords[1:-1]))}
        )

        cnt += 1

    if Point(graph.nodes[node_b]["geometry"].coords[0]).almost_equals(
        Point(line_b.coords[-1])
    ):
        if node_a == node_b and graph.has_edge(*(node_a, new_node_id, 0)):
            if line_b != graph.edges[(node_a, new_node_id, 0)]["geometry"]:
                k_new = 1
            else:
                k_new = 0
        else:
            k_new = 0
        edge_data.update(
            length=line_length(line_b, graph_crs),
            geometry=line_b,
            node_A=new_node_id,
            node_B=node_b,
            edge_fid=f"{new_node_id}_{node_b}",
        )
        graph.add_edge(
            new_node_id, node_b, k_new, **edge_data
        )  # changed the u, v order

        # Update the inverse vertices dict
        inverse_vertices_dict.update(
            {p: (node_b, new_node_id, k_new) for p in set(list(line_b.coords[1:-1]))}
        )

        cnt += 1

    if Point(graph.nodes[node_b]["geometry"].coords[0]).almost_equals(
        Point(line_a.coords[0])
    ):
        if node_a == node_b and graph.has_edge(*(node_a, new_node_id, 0)):
            if line_a != graph.edges[(node_a, new_node_id, 0)]["geometry"]:
                k_new = 1
            else:
                k_new = 0
        else:
            k_new = 0
        edge_data.update(
            length=line_length(line_a, graph_crs),
            geometry=line_a,
            node_A=new_node_id,
            node_B=node_b,
            edge_fid=f"{new_node_id}_{node_b}",
        )
        graph.add_edge(
            new_node_id, node_b, k_new, **edge_data
        )  # changed the u, v order

        # Update the inverse vertices dict
        inverse_vertices_dict.update(
            {p: (node_b, new_node_id, k_new) for p in set(list(line_a.coords[1:-1]))}
        )

        cnt += 1

    if Point(graph.nodes[node_a]["geometry"].coords[0]).almost_equals(
        Point(line_a.coords[-1])
    ):
        if node_a == node_b and graph.has_edge(*(node_a, new_node_id, 0)):
            if line_a != graph.edges[(node_a, new_node_id, 0)]["geometry"]:
                k_new = 1
            else:
                k_new = 0
        else:
            k_new = 0
        edge_data.update(
            length=line_length(line_a, graph_crs),
            geometry=line_a,
            node_A=node_a,
            node_B=new_node_id,
            edge_fid=f"{node_a}_{new_node_id}",
        )
        graph.add_edge(node_a, new_node_id, k_new, **edge_data)

        # Update the inverse vertices dict
        inverse_vertices_dict.update(
            {p: (node_a, new_node_id, k_new) for p in set(list(line_a.coords[1:-1]))}
        )

        cnt += 1

    if Point(graph.nodes[node_b]["geometry"].coords[0]).almost_equals(
        Point(line_a.coords[-1])
    ):
        if node_a == node_b and graph.has_edge(*(node_a, new_node_id, 0)):
            if line_a != graph.edges[(node_a, new_node_id, 0)]["geometry"]:
                k_new = 1
            else:
                k_new = 0
        else:
            k_new = 0
        edge_data.update(
            length=line_length(line_a, graph_crs),
            geometry=line_a,
            node_A=new_node_id,
            node_B=node_b,
            edge_fid=f"{new_node_id}_{node_b}",
        )
        graph.add_edge(
            new_node_id, node_b, k_new, **edge_data
        )  # changed the u, v order

        # Update the inverse vertices dict
        inverse_vertices_dict.update(
            {p: (node_b, new_node_id, k_new) for p in set(list(line_a.coords[1:-1]))}
        )

        cnt += 1

    if Point(graph.nodes[node_a]["geometry"].coords[0]).almost_equals(
        Point(line_a.coords[0])
    ):
        if node_a == node_b and graph.has_edge(*(node_a, new_node_id, 0)):
            if line_a != graph.edges[(node_a, new_node_id, 0)]["geometry"]:
                k_new = 1
            else:
                k_new = 0
        else:
            k_new = 0
        edge_data.update(
            length=line_length(line_a, graph_crs),
            geometry=line_a,
            node_A=node_a,
            node_B=new_node_id,
            edge_fid=f"{node_a}_{new_node_id}",
        )
        graph.add_edge(node_a, new_node_id, k_new, **edge_data)

        # Update the inverse vertices dict
        inverse_vertices_dict.update(
            {p: (node_a, new_node_id, k_new) for p in set(list(line_a.coords[1:-1]))}
        )

        cnt += 1

    try:
        assert cnt == 2
    except AssertionError:
        logging.warning(
            "No combination of nodes/road segments found for the OD assignment."
        )

    if graph.has_edge(node_a, node_b, k):
        # remove the edge that is split in two from the graph
        graph.remove_edge(node_a, node_b, k)

    return graph, inverse_vertices_dict


def add_od_nodes(
    od: gpd.GeoDataFrame,
    graph: nx.Graph,
    crs,
    category: Optional[str] = None,
):
    def find_closest_node(
        closest_node_on_road: np.ndarray,
        inverse_vertices_dict: dict,
        inverse_nodes_dict: dict,
        graph: nx.Graph,
    ) -> dict[tuple[float, float], int]:
        closest_u_v_k = inverse_vertices_dict.get(
            (closest_node_on_road[0], closest_node_on_road[1]), None
        )
        if closest_u_v_k:
            closest_u_data = graph.nodes[closest_u_v_k[0]]
            closest_v_data = graph.nodes[closest_u_v_k[1]]
            closest_node_on_extremities = closest_node(
                np.array((closest_node_on_road[0], closest_node_on_road[1])),
                np.array(
                    [
                        [closest_u_data["x"], closest_u_data["y"]],
                        [closest_v_data["x"], closest_v_data["y"]],
                    ]
                ),
            )
            closest_node_on_extremities_id = get_node_id_from_position(
                graph, *closest_node_on_extremities
            )
            inverse_nodes_dict[
                (closest_node_on_road[0], closest_node_on_road[1])
            ] = closest_node_on_extremities_id

        return inverse_nodes_dict

    def get_node_id_from_position(
        g: nx.Graph, x: float, y: float
    ) -> Union[float, None]:
        nodes = [
            (node, data)
            for node, data in g.nodes(data=True)
            if "geometry" in data
            and round(data["geometry"].x, 7) == round(x, 7)
            and round(data["geometry"].y, 7) == round(y, 7)
        ]
        return nodes[0][0] if nodes else None

    """Gets from each origin and destination the closest vertex on the graph edge.
    Args:
        od [Geodataframe]: The GeoDataFrame with the origins and destinations
        graph [networkX graph]: networkX graph
        crs [string or pyproj crs]:
    Returns:
        od [Geodataframe]: The GeoDataFrame with the locations of the origins and destinations on the road vertices
        graph [networkX graph]: The networkX graph updated with origins and destinations
    """
    logging.info("Finding vertices closest to Origins and Destinations")

    # create dictionary of the roads geometries and identifyers
    edge_list = [e for e in graph.edges.data(keys=True) if "geometry" in e[-1]]
    inverse_vertices_dict = {}
    all_vertices = []
    checked_lines = set()
    for line in edge_list:
        # Convert LineString to a hashable type (tuple) for set lookup
        geometry_coords = line[-1]["geometry"].coords
        coords = tuple(sorted([coord for coord in geometry_coords]))
        if coords in checked_lines:
            graph.remove_edge(*line[0:3])
        else:
            inverse_vertices_dict.update(
                {p: (line[0], line[1], line[2]) for p in set(geometry_coords[1:-1])}
            )
            all_vertices.extend(set(geometry_coords))
            checked_lines.add(coords)

    # Make an array from the list
    all_vertices = np.array(all_vertices)

    # Also create an inverse nodes dict of the node geometries as keys and node ID's as values
    inverse_nodes_dict = {
        node[-1]["geometry"].coords[0]: node[0] for node in graph.nodes.data()
    }

    # Get the maximum node id
    max_node_id = max([n for n in graph.nodes()])
    od_list = []
    for i, od_data in tqdm(
        enumerate(
            list(zip(od["geometry"].x, od["geometry"].y, od["o_id"], od["d_id"]))
        ),
        desc="Adding Origin-Destination nodes to graph",
    ):
        match_name = get_od(od_data[-2], od_data[-1])

        # Find the vertex on the road that is closest to the origin or destination point
        closest_node_on_road = closest_node(
            np.array((od_data[0], od_data[1])), all_vertices
        )
        match_od = Point(closest_node_on_road)
        # Find the road to which this vertex belongs. If the vertex is on an end-point of a road, it cannot be found,
        # and it goes to the except statement.
        try:
            closest_u_v_k = inverse_vertices_dict[
                (closest_node_on_road[0], closest_node_on_road[1])
            ]
            match_edge = graph.edges[closest_u_v_k]

            # Add the node to the graph edge
            match_geom = match_edge["geometry"]

            new_lines = split_line_with_points(match_geom, [match_od])
            if len(new_lines) == 1:
                inverse_nodes_dict = find_closest_node(
                    closest_node_on_road,
                    inverse_vertices_dict,
                    inverse_nodes_dict,
                    graph,
                )

            assert len(new_lines) == 2
            assert len([match_od]) == 1

            line1, line2 = new_lines

            new_node_id = max_node_id + 1
            max_node_id = new_node_id

            node_info = {
                "node_fid": new_node_id,  # Check if this attribute always exists
                "y": match_od.coords[0][1],
                "x": match_od.coords[0][0],
                "geometry": match_od,
                "od_id": match_name,
            }
            if category and od_data[-1] == od_data[-1]:
                # If the user wants to calculate the routes to multiple locations with categories
                # and if the current location is a destination (od_data[-1] is not NaN)
                node_info["category"] = od.iloc[i][category]

            # Add the new node to the graph
            graph.add_node(new_node_id, **node_info)

            # Update the inverse_nodes_dict with the new node
            inverse_nodes_dict[match_od.coords[0]] = new_node_id

            # Delete the new node from the inverse_vertices_dict as no end-points are included here
            del inverse_vertices_dict[match_od.coords[0]]

            # Update the graph edges and the inverse_vertices_dict
            graph, inverse_vertices_dict = update_edges_with_new_node(
                graph,
                match_edge,
                closest_u_v_k[0],
                closest_u_v_k[1],
                closest_u_v_k[2],
                line1,
                line2,
                new_node_id,
                crs,
                inverse_vertices_dict,
            )

        except (KeyError, AssertionError):
            # If the vertex is at the end of the road it won't be found in the inverse_vertices_dict,
            # so search in the inverse_nodes_dict.
            match_node = inverse_nodes_dict[
                (closest_node_on_road[0], closest_node_on_road[1])
            ]

            # Update the node with the OD attribute
            graph = add_data_to_existing_node(graph, match_node, match_name)

            if category and od_data[-1] == od_data[-1]:
                # If the user wants to calculate the routes to multiple locations with categories
                # and if the current location is a destination (od_data[-1] is not NaN)
                graph.nodes[match_node]["category"] = od.iloc[i][category]

        # Save both in lists
        od_list.append(match_od)  # save the point as a Shapely Point

    # save in dataframe
    od["OD"] = od_list

    # save the road vertices closest to the origin/destination as geometry, delete the input origin/destination point geometry
    od = gpd.GeoDataFrame(od)
    od = od.drop(columns=["OD"])

    return od, graph


def split_line_with_points(line, points):
    """Splits a line string in several segments considering a list of points."""
    segments = []
    current_line = line

    # make a list of points and its distance to the start to sort them from small to large distance
    list_dist = [current_line.project(pnt) for pnt in points]
    list_dist.sort()

    for d in list_dist:
        # cut the line at a distance d
        seg, current_line = cut(current_line, d)
        if seg:
            segments.append(seg)
    segments.append(current_line)
    return segments


#########################################################################################
################### Code to generate origins points from raster #########################
#########################################################################################


def rescale_and_crop(path_name, gdf, output_folder: Path, res: int = 500):
    dst_crs = rasterio.crs.CRS.from_dict(gdf.crs.to_dict())

    # Rescale and reproject raster to gdf crs
    _output_origins_raster_tif = output_folder / "origins_raster_reprojected.tif"

    with rasterio.open(path_name) as src:
        transform, width, height = calculate_default_transform(
            src.crs, dst_crs, src.width, src.height, *src.bounds
        )

        m2degree = (
            1 / 111000
        )  # approximate conversion from meter to 1 degree of EPSG:4326; TODO: make flexible depending on the input crs
        transform = Affine(
            res * m2degree,
            transform.b,
            transform.c,
            transform.d,
            -res * m2degree,
            transform.f,
        )

        # use scale, instead of absolute meter, for resolution
        # scale = 2
        # transform = Affine(transform.a * scale, transform.b, transform.c, transform.d, transform.e * scale, transform.f)
        # height = height * scale
        # width = width * scale

        kwargs = src.meta.copy()
        kwargs.update(
            {"crs": dst_crs, "transform": transform, "width": width, "height": height}
        )

        with rasterio.open(_output_origins_raster_tif, "w", **kwargs) as dst:
            for i in range(1, src.count + 1):
                reproject(
                    source=rasterio.band(src, i),
                    destination=rasterio.band(dst, i),
                    src_transform=src.transform,
                    src_crs=src.crs,
                    dst_transform=transform,
                    dst_crs=dst_crs,
                    resampling=Resampling.sum,
                )  # Resampling.sum or Resampling.nearest

    raster = rasterio.open(_output_origins_raster_tif)

    # Crop to shapefile
    out_array, out_trans = rasterio.mask.mask(
        dataset=raster, shapes=gdf.geometry, crop=True
    )

    out_meta = raster.meta.copy()
    out_meta.update(
        {
            "height": out_array.shape[1],
            "width": out_array.shape[2],
            "transform": out_trans,
        }
    )
    raster.close()
    os.remove(_output_origins_raster_tif)

    return out_array, out_meta


def export_raster_to_geotiff(array, meta, dir_path: Path, filename: str) -> Path:
    cropped_outputfile = dir_path / filename
    with rasterio.open(
        cropped_outputfile, "w", **meta, compress="LZW", tiled=True
    ) as dest:
        dest.write(array)
    return cropped_outputfile


def generate_points_from_raster(fn, out_fn):
    # Read raster coordinate centroid
    with rasterio.open(fn) as src:
        points = []
        for col in range(src.width):
            x_s, y_s = rasterio.transform.xy(
                src.transform,
                [row for row in range(src.height)],
                [col for _ in range(src.height)],
            )
            points += [Point(x, y) for x, y in zip(x_s, y_s)]

    # Put raster coordinates into geodataframe
    gdf = gpd.GeoDataFrame()
    gdf["geometry"] = points
    gdf["OBJECTID"] = [x for x in range(len(gdf))]

    # Get raster values
    temp = [gdf["geometry"].x, gdf["geometry"].y]
    coords = list(map(list, zip(*temp)))
    with rasterio.open(fn) as src:
        gdf.crs = src.crs
        gdf["values"] = [sample[0] for sample in src.sample(coords)]

    # Save non-zero cells
    gdf.loc[gdf["values"] > 0].to_file(out_fn)

    return out_fn


def origins_from_raster(output_folder: Path, mask_fn, raster_fn) -> Path:
    """Makes origin points from a population raster."""
    output_fn = output_folder / "origins_raster.tif"
    mask = gpd.read_file(mask_fn[0], engine="pyogrio")
    res = 1000  # in meter; TODO: put in config file or in network.ini
    out_array, out_meta = rescale_and_crop(raster_fn, mask, output_folder, res)
    outputfile = export_raster_to_geotiff(out_array, out_meta, output_folder, output_fn)

    out_array[out_array > 0] = 1
    print(
        "There are "
        + str(out_array[~np.isnan(out_array)].sum().sum())
        + " origin points."
    )

    out_fn = output_folder / "origins_points.gpkg"
    out_fn = generate_points_from_raster(outputfile, out_fn)

    return out_fn



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007
    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from pathlib import Path

import pandas as pd

from ra2ce.common.io.writers.ra2ce_exporter_protocol import Ra2ceExporterProtocol
from ra2ce.network.avg_speed.avg_speed import AvgSpeed


class AvgSpeedWriter(Ra2ceExporterProtocol):
    def export(self, export_path: Path, export_data: AvgSpeed) -> None:
        export_path.parent.mkdir(parents=True, exist_ok=True)
        pd.DataFrame(
            {
                "road_types": list(map(str, export_data.speed_per_road_type.keys())),
                "avg_speed": export_data.speed_per_road_type.values(),
            }
        ).to_csv(export_path)
